"""Utitlities."""

import re
from typing import List, Tuple, Optional

import requests
from bs4 import BeautifulSoup
from requests import Response


class UrlData:
    """Url stuff."""

    _resp: Optional[Response]
    _html: Optional[str]
    _bs: Optional[BeautifulSoup]

    def __init__(self, url: str):
        """Build url."""
        self.headers: dict = {}

        self.html_parser: str = "lxml"

        self._url = url
        self._resp = None
        self._html = None
        self._bs = None

    def __str__(self):
        """Return string rep."""
        return "<{}>".format(self.url)

    @property
    def resp(self) -> Response:
        """Get the requests response object."""
        if not self._resp:
            self._resp = requests.get(self.url, headers=self.headers)
        return self._resp

    @property
    def html(self) -> str:
        """Get the html for this url."""
        if not self._html:
            self._html = self.resp.text
        return self._html

    @property
    def bs(self) -> BeautifulSoup:
        """Get the BeautifulSoup object."""
        if not self._bs:
            self._bs = BeautifulSoup(self.html, self.html_parser)
        return self._bs

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, value: str):
        self._url = value
        self._resp = None
        self._html = None
        self._bs = None


def search(query: str, api_key: str) -> List:
    """Return search results."""

    # ! to edit head to: https://cse.google.com/cse/all

    params = {
        "key": api_key,
        #"cx": "'000281471148392423350:ymsqkb0dqs8"
        "cx": "000281471148392423350:64fnnp-ny2w",  # original + darklyrics
        #"cx": "002017775112634544492:7y5bpl2sn78", # original without darklyrics
        "q": query
    }
    resp = requests.get("https://www.googleapis.com/customsearch/v1",
                        params=params)
    data = resp.json()
    items = data.get("items", [])
    return items


def generate_url(artist: str, album: str, song: str) -> List[dict]:

    # the list of words that are not capitalized, may not be exhaustive!!
    DONT_CAPITALIZE: Tuple[str, ...] = ("on", "at", "by", "or", "a", "an",
                                        "of", "and", "but")

    # no = ["under","between", "over", "after","without","until",
    #       "because","every","this","those", "many", ]
    # maybe = ["and","but",]

    def r(string: str, replace: str) -> str:
        return string.replace(" ", replace)

    urls: List[dict] = []

    artist = re.sub(r"\(|\)|\'", "", artist).strip().lower()
    album = re.sub(r"\(|\)|\'", "", album).strip().lower()
    song = re.sub(r"\(|\)|\'", "", song).strip().lower()

    # darklyrics
    urls.append({
        "link":
        f"http://www.darklyrics.com/lyrics/{r(artist, '')}/" +
        f"{r(album, '')}.html"
    })
    # AZLyrics
    urls.append({
        "link":
        f"https://www.azlyrics.com/lyrics/{r(artist, '')}/" +
        f"{r(song, '')}.html"
    })
    # Genius
    link = (f"https://genius.com/{r(artist.capitalize(), '-')}"
            f"-{r(song, '-')}-lyrics").replace("&", "and")
    urls.append({"link": link})  # sometimes annotated version is present
    urls.append({"link": f"{link}-annotated"})
    # Lyricsmode
    urls.append({
        "link":
        f"http://www.lyricsmode.com/lyrics/{artist[0]}/" +
        f"{r(artist, '_')}/{r(song, '_')}.html"
    })
    # Musixmatch
    song = "-".join([
        s.capitalize() for s in song.split() if s not in DONT_CAPITALIZE
    ]).capitalize()
    urls.append({
        "link":
        f"https://www.musixmatch.com/lyrics/" +
        f"{r(artist.upper(), '-')}/{song}"
    })
    # Lyrical Nonsense, Anime Lyrics - useless, some Japanese lyrics

    return urls


def safe_filename(name: str, file_ending: str = ".json") -> str:
    """Return a safe version of name + file_type."""
    filename = re.sub(r"\s+", "_", name)
    filename = re.sub(r"\W+", "-", filename)

    return filename.lower().strip() + file_ending


def clean_lyrics(lyrics: str) -> str:
    """Perform some simple operations to clean the lyrics."""
    lyrics = lyrics.strip()
    # remove unwanted characters
    lyrics = re.sub(r"[^\w\[\]()/ \"',\.:\-\n?!]+", "", lyrics)
    # reduce to one space only
    lyrics = re.sub(r" +", " ", lyrics)
    # reduce to max 2 new lines in a row
    lyrics = re.sub(r"\n{2,}", "\n\n", lyrics)
    # remove space before newline
    lyrics = re.sub(r" +?\n", "\n", lyrics)

    return lyrics



"""Lyric finder exceptions."""

LyricsException = type("LyricsException", (Exception,), {})
NoLyrics = type("NoLyrics", (LyricsException,), {})
NotAllowedError = type("NotAllowedError", (LyricsException,), {})


class NoExtractorError(LyricsException):
    """When there's no extractor for a url."""

    def __init__(self, url):
        """Create new."""
        super().__init__("No extractor found for {}".format(url))



"""
Get lyrics from:

Anime Lyrics, AZLyrics, Genius, Lyricsmode, \
Lyrical Nonsense, Musixmatch, darklyrics
"""

import logging
from typing import TYPE_CHECKING, Dict, List, Optional, Union

import fuzzywuzzy.fuzz as fuzz  # lazy loaded

from wiki_music.constants import GREEN, NO_LYRIS, RESET
from wiki_music.external_libraries import lyricsfinder  # lazy loaded
from wiki_music.utilities import (SharedVars, ThreadPool, caseless_equal,
                                  exception, normalize, read_google_api_key,
                                  we_are_frozen)

if TYPE_CHECKING:
    from wiki_music.external_libraries.lyricsfinder import LyricsManager

log = logging.getLogger(__name__)
log.debug("lyrics imports done")

__all__ = ["save_lyrics"]


def save_lyrics(tracks: List[str], types: List[str], band: str, album: str,
                GUI: bool) -> List[str]:
    """ This function does some preprocessing before it starts the lyricsfinder
    module and downloads the lyrics. In preproces, tracks which will have same
    lyrics are identified so the same lyrics are not downloaded twice. The
    lyrics are then downloaded asynchronously each in separate thread for
    speed.

    See also
    --------
    :mod:`wiki_music.external_libraries.lyricsfinder`
         module used to download lyrics
    :class:`wiki_music.utilities.parser_utils.ThreadPool`
        async download

    Parameters
    ----------
    tracks: List[str]
        list of album tracks
    types: List[str]
        list of album types, to infer which tracks have same lyrics, e.g. 
        <track> and <track (acoustic)> are considered to have same lyrics.
        Instrumental and Orchestral types are set to no lyrics
    band: str
        album artist name
    album: str
        album name

    Returns
    -------
    List[str]
        list of track lyrics in same order as tracks list was passed in
    """

    GOOGLE_API_KEY = read_google_api_key(GUI)

    log.info("starting save lyrics")

    lyrics: List[str]
    tracks_dict: Dict[str, Dict[str, Union[str, list]]]
    raw_lyrics: List[Dict[str, Optional[str]]]

    lyrics = []
    for i, tp in enumerate(types):
        for nl in NO_LYRIS:
            if caseless_equal(nl, tp):
                lyrics.append(nl)
                break
        else:
            lyrics.append("")

    log.info("Initialize duplicates")

    tracks_dict = dict()

    for i, (tr, lyr) in enumerate(zip(tracks, lyrics)):

        if not lyr:
            for tr_k in tracks_dict.keys():
                if fuzz.token_set_ratio(tr, tr_k) > 90:
                    tracks_dict[tr_k]["track"].append(i)
                    break
            else:
                tracks_dict[tr] = {"track": [i]}

    log.info("Download lyrics")

    # manager must be initialized in main thread
    manager = lyricsfinder.LyricsManager()

    # adjust logging level for frozen app
    if we_are_frozen():
        name = "wiki_music.external_libraries.lyricsfinder.lyrics"
        logging.getLogger(name).setLevel("ERROR")

    # run search
    raw_lyrics = ThreadPool(target=_get_lyrics,
                            args=[(manager, band, album, t, GOOGLE_API_KEY)
                                  for t in tracks_dict.keys()]).run()

    log.info("Assign lyrics to tracks_dict")

    # report results
    for i, l in enumerate(raw_lyrics):
        if l["lyrics"]:
            print(GREEN + "Saved lyrics for:" + RESET,
                  f"{l['artist']} - {l['title']} " + GREEN +
                  f"({l['origin']['source_name']})")
            tracks_dict[l["title"]]["lyrics"] = l["lyrics"]
        else:
            print(GREEN + "Couldn't find lyrics for:" + RESET,
                  f"{l['artist']} - {l['title']}")
            tracks_dict[l["title"]]["lyrics"] = ""

    for track in tracks_dict.values():
        for t in track["track"]:
            lyrics[t] = track["lyrics"]

    return lyrics


@exception(log)
def _get_lyrics(manager: 'LyricsManager', artist: str, album: str, song: str,
                GOOGLE_API_KEY: str
                ) -> Dict[str, Optional[Union[str, Dict[str, str]]]]:
    """ Function which calls lyricsfinder.LyricsManager.search_lyrics method
    to find and download lyrics for specified song.

    See also
    --------
    :meth:`wiki_music.external_libraries.lyricsfinder.LyricsManager`

    Parameters
    ----------
    manager: lyricsfinder.LyricsManager
        instance of LyricsManager which encapsulates the lyrics finding and
        downloading methods
    artist: str
        artist name
    album: str
        album name
    song: str
        song name

    Returns
    -------
    dict
        dictionary with lyrics and information where it was downloaded from
    """

    lyrics = next(
        manager.search_lyrics(song,
                              album,
                              artist,
                              google_api_key=GOOGLE_API_KEY), None)

    if not lyrics:
        log.info(f"Couldn't find lyrics for: {artist} - {song}")
        return {"lyrics": None, "artist": artist, "title": song}
    else:
        log.info(f"Saved lyrics for: {artist} - {song}")
        l = lyrics.to_dict()
        l["title"] = song
        l["lyrics"] = normalize(l["lyrics"].replace("\r",
                                                    "").replace("\n", "\r\n"))

        return l



""" Base module for all tag handlers. """

import collections
import logging
from abc import ABC, abstractmethod
from typing import ClassVar, Dict, Optional

logging.getLogger(__name__)


class SelectiveDict(dict):
    """A subclass of a dictionary with capacity to remember which keys are
    being changed. This is archieved by simply overriding the __setitem__
    method.

    Attributes
    ----------
    writable: set
        a set of keys(tags) that were changed, and should be written to file
    """

    def __init__(self, *args):
        super().__init__(*args)
        self.writable = set()

    def __getitem__(self, key):
        val = super().__getitem__(key)
        return val

    def __setitem__(self, key, val):
        """Altered magic method which caches the changed keys names.
        
        Parameters
        ----------
        key: str
            dictionary key (tag name)
        val: Any
            dictionary value (tag value)
        """

        try:
            old_val = super().__getitem__(key)
        except KeyError:
            self.writable.add(key)
        else:
            if old_val != val:
                self.writable.add(key)
        finally:
            super().__setitem__(key, val)

    def save_items(self):
        """Method simillar to dict.items() method only this one loops
        through key-value pairs that have changed since the creation of
        this dictionary instance.
        """

        for key, val in super().items():
            if key in self.writable:
                yield key, val


class TagBase(ABC):
    """

    Attributes
    ----------
    _map_keys: dict
        a maping between highlevel tag names and lowlevel tag names used by
        mutagen impementation of tags for specific file type
    _reverse_map: dict
        reversed `_map_keys` dictionary
    _tags: Selective_dict
        private attribute which caches the read tags and records any occuring
        changes, it is exposed in class API through tags property

    See also
    --------
    :class:`wiki_music.library.tags_handler.tag_base.SelectiveDict`
        a subclass of a dictionary in which tags are stored, does remember
        changes in tags, so only the changed tags can be written
    :const:`wiki_music.constants.tags.TAGS`
        for list of supported tags

    Parameters
    ----------
    filename: str
        path to song file with tags
    """

    __doc__: Optional[str]
    _map_keys: ClassVar[Dict[str, str]]
    reverse_map: Dict[str, str]
    _tags: SelectiveDict

    def __init__(self, filename):

        self._song = None
        self._tags = None

        self.reverse_map = self._get_reversed(self._map_keys)

        self._open(filename)

    def save(self):
        """Writes tags to song file and than calls the save function, to save
        to disk.
        """

        for tag, value in self._tags.save_items():
            self._write(tag, value)

        self._song.save()

    @abstractmethod
    def _read(self):
        """Reads tags from an open mutagen file to dictionary as key-value
        pairs. Tries to avoid all the pitfalls of different tag formats.

        Returns
        -------
        dict
            dictionary of tag names and values
        """
        raise NotImplementedError("Call to abstarct method!")

    @abstractmethod
    def _write(self, tag, value):
        """Given a high level tag name, convert it to low level name and write
        to file tags."""
        raise NotImplementedError("Call to abstarct method!")

    @abstractmethod
    def _open(self, filename):
        """Method that reads in the file frim location suplied by filename
        argument. see subclasses for specific implementation.
        """
        raise NotImplementedError("Call to abstarct method!")

    @property
    def tags(self):
        """Propery which is used to access the read tags. If the tags are not
        present it reads them from a suplied file and casts them from
        dictionary to SelectiveDict type to record occured changes

        Returns
        -------
        SelectiveDict
            dictionary containing tag labes and their values
        """

        if not self._tags:
            self._tags = self._read()

        # for reading tags we use standard dict but we expose to outer
        # world the SelectiveDict class to record occured changes
        if not isinstance(self._tags, SelectiveDict):
            self._tags = SelectiveDict(self._tags)

        return self._tags

    @staticmethod
    def _get_reversed(_map_keys: Dict[str, str]) -> Dict[str, str]:
        """Given a dictionary of [keys, values] it returns a reversed version
        with [values, key] while preserving order of items if is an instance
        of collections.OderedDict.

        Returns
        -------
        dict
            dictionary with switched keys and values
        """

        reverse_map: Dict[str, str] = collections.OrderedDict()

        for key, value in _map_keys.items():
            reverse_map[value] = key

        return reverse_map



"""Module that manages GUI data model and interacion with parser.

Warnings
--------
All parser interaciton should be hanhled by this class. Parser methods and
attributes should not be accesed directly in GUI.
""" 

import logging
from typing import Iterable, List, Optional, Union

from wiki_music.constants import GUI_HEADERS, SPLIT_HEADERS, STR_TAGS
from wiki_music.gui_lib import (BaseGui, CustomQStandardItem,
                                CustomQStandardItemModel, NumberSortModel,
                                ResizablePixmap)
from wiki_music.gui_lib.qt_importer import (QImage, QLabel, QModelIndex,
                                            QPixmap, QStandardItemModel,
                                            QTimer)
from wiki_music.library.parser import WikipediaRunner
from wiki_music.utilities import SharedVars, exception

log = logging.getLogger(__name__)
log.debug("data model imports done")


class ParserInteract(BaseGui):
    """Class that represents comunication chanel between parser and GUI.

    Warnings
    --------
    This class is not ment to be instantiated, only inherited.

    Attributes
    ----------
    _parser: :class:`wiki_music.library.parser.WikipediaRunner`
        parser instance doing all the heavy lifting
    """

    def __init__(self) -> None:

        self._parser = WikipediaRunner(GUI=True)
        super().__init__()

        # TODO needs QThreads to work
        # self._parser.selectedGenreChanged.connect(self.genre_entry.setText)
        # self._parser.releaseDateChanged.connect(self.year_entry.setText)
        # self._parser.bandChanged.connect(self.band_entry)
        # self._parser.bandChanged.connect(self.band_entry_input)
        # self._parser.albumChanged.connect(self.album_entry)
        # self._parser.albumChanged.connect(self.album_entry_input)

    @property
    def GENRE(self) -> str:
        """Found genre.

        See also
        --------
        :attr:`wiki_music.library.parser.base.ParserBase.GENRE`
            parser attribute tied to this property
        :class:`BaseGui.genre_entry`
            Qt entry field tied to this property

        :type: str
        """
        return self._parser.selected_genre

    @GENRE.setter
    def GENRE(self, value: str):
        self._parser.selected_genre = value
        self.genre_entry.setText(value)

    @property
    def DATE(self) -> str:
        """Found release date.

        See also
        --------
        :attr:`wiki_music.library.parser.base.ParserBase.DATE`
            parser attribute tied to this property
        :class:`BaseGui.year_entry`
            Qt entry field tied to this property

        :type: str
        """
        return self._parser.release_date

    @DATE.setter
    def DATE(self, value: str):
        self._parser.release_date = value
        self.year_entry.setText(value)

    @property
    def ALBUM(self) -> str:
        """Found album name.

        See also
        --------
        :attr:`wiki_music.library.parser.base.ParserBase.ALBUM`
            parser attribute tied to this property
        :class:`BaseGui.album_entry`
            Qt entry field tied to this property
        :class:`BaseGui.album_entry_input`
            Qt entry field tied to this property

        :type: str
        """
        return self._parser.album

    @ALBUM.setter
    def ALBUM(self, value: str):
        self._parser.album = value
        self.album_entry.setText(value)
        self.album_entry_input.setText(value)

    @property
    def ALBUMARTIST(self) -> str:
        """Found album artist name.

        See also
        --------
        :attr:`wiki_music.library.parser.base.ParserBase.ALBUMARTIST`
            parser attribute tied to this property
        :class:`BaseGui.band_entry`
            Qt entry field tied to this property
        :class:`BaseGui.band_entry_input`
            Qt entry field tied to this property

        :type: str
        """
        return self._parser.band

    @ALBUMARTIST.setter
    def ALBUMARTIST(self, value: str):
        self._parser.band = value
        self.band_entry.setText(value)
        self.band_entry_input.setText(value)

    @property
    def url(self) -> str:
        """Wikipedia url address of the downloaded page.

        See also
        --------
        :attr:`wiki_music.library.parser.preload.WikiCooker.url`
            parser attribute tied to this property

        :type: str
        """
        return self._parser.url

    @property
    def genres(self) -> List[str]:
        """List of genres found on wikipedia.

        See also
        --------
        :attr:`wiki_music.library.parser.base.ParserBase.genres`
            parser attribute tied to this property

        :type: str
        """
        return self._parser.genres

    @property  # type: ignore
    def work_dir(self) -> str:  # type: ignore
        """Current working directory.

        See also
        --------
        :attr:`wiki_music.library.parser.base.ParserBase.work_dir`
            parser attribute tied to this property
        :class:`BaseGui.display_dir`
            Qt entry field tied to this property

        :type: str
        """
        return self._parser.work_dir

    @work_dir.setter
    def work_dir(self, value: str):
        self.display_dir.setText(value)
        self._parser.work_dir = value

    @property
    def COVERART(self):
        """Album coverart.

        See also
        --------
        :attr:`wiki_music.library.parser.base.ParserBase.COVERART`
            parser attribute tied to this property

        :type: str
        """
        return self._parser.COVERART

    @COVERART.setter
    def COVERART(self, value: bytearray):
        self._parser.COVERART = value

    def stop_preload(self):
        """Stops running page preload.

        See also
        --------
        :class:`wiki_music.library.parser.preload.WikiCooker.Preload`
            parser inner class controlling preload
        """
        self._parser.Preload.stop()

    def start_preload(self):
        """Starts page preload.

        See also
        --------
        :class:`wiki_music.library.parser.preload.WikiCooker.Preload`
            parser inner class controlling preload
        """
        self._parser.Preload.start()

    def write_tags(self) -> bool:
        """Writes tags to music files.

        See also
        --------
        :class:`wiki_music.library.parser.in_out.ParserInOut.write_tags`
            method that controls tag writing

        Returns
        -------
        bool
            truth if writing was successful
        """
        self._gui_to_parser()
        return self._parser.write_tags()

    def read_files(self):
        """Rears tags from music files.

        See also
        --------
        :class:`wiki_music.library.parser.in_out.ParserInOut.read_files`
            method for reading tags from files
        """
        self._parser.read_files()

    @property
    def number_of_tracks(self):
        """Number of found tracks.

        Number of tracks detected by parser on wikipedia or load from disk.

        :type: int            
        """
        return len(self._parser)

    @property
    def save_ready(self) -> bool:
        """Tells if there are any writable files in working directory.

        :type: bool
        """
        return bool(self._parser)

    def reinit_parser(self):
        """Sets parser and SharedVars attributes to their default values.

        Reinitializes parser and SharedVars attributes so a new search can be
        performed.
        """

        # TODO non-atomic
        self._parser.__init__(protected_vars=False)
        self._parser.list_files()
        SharedVars.re_init()


class DataModel(ParserInteract):
    """Transfer data between GUI and parser and manage GUI data model.

    This class reads data from parser to be displayed in GUI and on save or
    other action that requires parser it collects data in GUI and writes it to
    parser.

    Warnings
    --------
    This class is not ment to be instantiated, only inherited.

    Attributes
    ----------
    cover_art: Optional[ResizablePixmap]
        Qt class that displays the found cover art picture
    table: CustomQStandardItemModel
        Qt table for displaying parser data
    proxy: NumberSortModel
        table model for data sorting
    """

    cover_art: Optional[ResizablePixmap]
    table: CustomQStandardItemModel
    proxy: NumberSortModel

    def __init__(self) -> None:

        log.debug("init data model")

        super().__init__()

        self.cover_art = None

        # create table with headers
        self.table = CustomQStandardItemModel()
        self.table.setHorizontalHeaderLabels(GUI_HEADERS)

        self.proxy = NumberSortModel()
        self.proxy.setSourceModel(self.table)

        log.debug("init data model done")

    def _gui_to_parser(self):
        """Transfers data from GUI to parser.

        Warnings
        --------
        This operation is non-atomic, as of now it still does not have a proper
        locking mechanism. We rely on the fact that only GUI or the parser
        are trying to access data at a given time.
        """

        col: int

        # TODO non-atomic
        for h in GUI_HEADERS:

            col = self.table[h]
            split = h in SPLIT_HEADERS

            col_data = [self.table.item(row, col).text(split=split)
                        for row in range(self.table.rowCount())]

            setattr(self._parser, h, col_data)

    def _display_image(self, image: Optional[bytearray] = None):
        """Shows cover art image preview in main window.

        See also
        --------
        :class:`wiki_music.gui_lib.custom_classes.ResizablePixmap`
        
        Parameters
        ----------
        image: bytearray
            force displaying of the input image instead of the one contained in
            parser. Also the input image is saved to parser.
        """

        if image:
            self.COVERART = image

        if self.cover_art:
            self.cover_art.update_pixmap(self.COVERART)
        else:
            self.cover_art = ResizablePixmap(self.COVERART)
            self.picture_layout.addWidget(self.cover_art)

    @exception(log)
    def _parser_to_gui(self):
        """Transfers data from parser GUI.

        Warnings
        --------
        This operation is non-atomic, as of now it still does not have a proper
        locking mechanism. We rely on the fact that only GUI or the parser
        are trying to access data at a given time.
        """
        col: int

        if self.number_of_tracks:

            log.debug("init entries")

            self.table.setRowCount(self.number_of_tracks)

            # TODO until QThreads are implemented and these can be passed
            # with signals
            for s in STR_TAGS:
                setattr(self, s, getattr(self._parser, s))

            # TODO non-atomic
            for idx, h in enumerate(GUI_HEADERS):
                col = self.table[h]
                for row, value in enumerate(getattr(self._parser, h)):
                    self.table.setItem(row, col, CustomQStandardItem(value))

                if idx != len(GUI_HEADERS) - 2:
                    self.tableView.resizeColumnToContents(col)

            self.tableView.update()

            log.debug("show cover art")
            self._display_image()

            log.debug("init entries done")
        else:
            QTimer.singleShot(100, self._parser_to_gui)

    @exception(log)
    def _detail(self, proxy_index: QModelIndex):
        """Display detail of the row that is clicked and sync changes to table.

        Parameters
        ----------
        proxy_index: QModelIndex
            index of the proxy model cell, needs to be translated to real table
            cell index
        """

        # proxy_index is QModelIndex proxyIndex Class type
        # it is index of data mapping as shown in gui
        # real indeces are different
        row = self.proxy.mapToSource(proxy_index).row()

        # maintain row highlight
        index = self.tableView.selectionModel().selectedRows()
        self.tableView.selectRow(index[0].row())

        entries = [
            (0, self.number_detail),
            (1, self.title_detail),
            (3, self.artist_detail),
            (4, self.composer_detail),
            (6, self.lyrics_detail),
            (7, self.file_detail)
        ]

        log.debug("connect detail function")

        for col, ent in entries:
            ent.disconnect()
            ent.setText(self.table.item(row, col).text())
            # col_inner = col is a hack to change the
            # scope of col otherwise lambda will refer
            # to actual value of col since it is in
            # outer scope
            # None is the default argument for text since QTextEdit
            # signal doesn´t retutn text
            ent.textChanged.connect(
                lambda text=None, c=col: self.__text_check__(row, c, text))

    def __text_check__(self, row: int, col: int, text: Optional[str]):
        """Writes changed text in detail tab to table.
        
        Parameters
        ----------
        row: int
            table row
        col: int
            table column
        text: Optional[str]
            text put in cell, if None read it from :attr:`lyrics_detail`
        """

        if text is None:
            text = self.lyrics_detail.toPlainText()

        self.table.setItem(row, col, CustomQStandardItem(text))



""" Constants used by whole :mod:`wiki_music` module."""
from typing import Tuple

__all__ = ["TAGS", "EXTENDED_TAGS", "STR_TAGS"]

#: enumeration of tags that we are able to read and write to music files
TAGS: Tuple[str, ...] = ("ALBUM", "ALBUMARTIST", "ARTIST", "COMPOSER",
                         "COVERART", "DATE", "DISCNUMBER", "GENRE", "LYRICS",
                         "TITLE", "TRACKNUMBER")
#: tags extended with file path and track type which is not directly
#: writen into file tags
EXTENDED_TAGS: Tuple[str, ...] = TAGS + ("FILE", "TYPE")
#: marks tags which have separate entry fields in gui and are not displayed
#: in tracklist table
STR_TAGS: Tuple[str, ...] = ("GENRE", "ALBUM", "ALBUMARTIST", "DATE")



"""Module handling cover art downloades from goole image search."""
from __future__ import absolute_import


def main():
    from google_images_download import google_images_download
    from google_images_download import google_images_download_offline

if __name__ == '__main__':
    main()



""" Base module for all parser classes from which they import ParserBase
class that sets all the default attributes.
"""

import logging
from typing import TYPE_CHECKING, Dict, List, Optional

from wiki_music.utilities import MultiLog

__all__ = ["ParserBase"]

log = logging.getLogger(__name__)

# only for static typechecker, is False at runtime
if TYPE_CHECKING:
    from wikipedia import WikipediaPage
    from bs4 import BeautifulSoup

    Bs4Soup = Optional["BeautifulSoup"]
    WikiPage = Optional["WikipediaPage"]

SList = List[str]  # list of strings
IList = List[int]  # list of ints
NSList = List[SList]  # nested list
NIList = List[IList]  # nested list


class ParserBase:
    """The base clas for all :mod:`wiki_music.parser` subclasses. Defines the
    necessary attributes. The uppercased attributes correspond to tag names
    for easier access.

    Warnings
    --------
    This class is not ment to be instantiated, only inherited.

    Note
    ----
    Uppercased propertie corespond with TAG names so we can easilly use getattr
    and setattr methods

    Attributes
    ----------
    contents: List[str]
        stores the wikipedia page contents
    disk_sep: List[int]
        list of tracks separating disks e.g. if CD 1 = (1, 13) and
        CD 2 = (14, 20), disk_sep = [0, 12, 19] the offset by one if because of
        zero first index
    disks: List[list]
        holds album disks titles
    genres: List[str]
        list of genres found in wikipedia page
    header: List[str]
        tracklist table headers
    NLTK_names: List[str]
        list of Person Named Entities extracted from wikipedia page by nltk.
        See :meth:`wiki_music.library.parser.process_page.WikipediaParser.extract_names`
        for details on how and from which parts of test the names are extracted
    personnel: List[str]
        list holding adittional personnel participating on album
    appearences: List[List[int]]
        list coresponding to personnel holding for each person list of tracks
        that the said person has appeared on
    subtracks: List[List[str]]
        each entry holds list of subtracks for one track
    sub_types: List[List[str]]
        each entry holds list of types for each subtrack
    work_dir: str
        string with path to directory with music files, this variable can be
        protected from reseting in __init__ method
    log: :class:`wiki_music.utilities.utils.MultiLog`
        instance of MUltiLog which sends messages to logger and
        :class:`wiki_music.utilities.sync.SharedVars`
    sections: Dict[str, List[Bs4Soup]]
        dictionary of lists of BeautifulSoup objects each entry in the
        dict contains one whole section of the page and is indexed by that
        section title
    """

    files: SList
    bracketed_types: SList
    sections: Dict[str, List["Bs4Soup"]]
    page: "WikiPage"
    soup: "Bs4Soup"

    def __init__(self, protected_vars: bool) -> None:

        log.debug("parser base")

        # lists 1D
        self.contents: SList = []
        self.tracks: SList = []
        self.types: SList = []
        self.disc_num: IList = []
        self.disk_sep: IList = []
        self.disks: List[list] = []
        self.genres: SList = []
        self.header: SList = []
        self.lyrics: SList = []
        self.NLTK_names: SList = []
        self.numbers: SList = []
        self.personnel: SList = []
        self._bracketed_types: SList = []
        self._files: SList = []

        # lists 2D
        self.appearences: NIList = []
        self.artists: NSList = []
        self.composers: NSList = []
        self.sub_types: NSList = []
        self.subtracks: NSList = []

        # bytearray
        self.cover_art: bytearray = bytearray()

        # strings
        self.release_date: str = ""
        self.selected_genre: str = ""

        # atributes protected from GUIs reinit method
        # when new search is started
        if protected_vars:
            self.album: str = ""
            self.band: str = ""
            self.work_dir: str = ""
            self.log: MultiLog = MultiLog(log)
            self.GUI = False

        self.log.debug("parser base done")

    def __len__(self):
        return len(self.numbers)

    def __bool__(self):
        return bool(self.__len__())

    @property
    def ALBUM(self) -> str:
        """ string with album name, this variable can be protected from
        reseting in __init__ method.
        
        :type: str
        """
        return self.album

    @ALBUM.setter
    def ALBUM(self, value: str):
        self.album = value

    @property
    def ALBUMARTIST(self) -> str:
        """ string with band name, this variable can be protected from reseting
        in __init__ method.

        :type: str
        """
        return self.band

    @ALBUMARTIST.setter
    def ALBUMARTIST(self, value: str):
        self.band = value

    @property
    def ARTIST(self) -> NSList:
        """ Each entry in list holds list of artists for one track.

        :type: List[List[str]]
        """
        return self.artists

    @ARTIST.setter
    def ARTIST(self, value: NSList):
        self.artists = value

    @property
    def COMPOSER(self) -> NSList:
        """ Each entry in list holds list of composers for one track.

        :type: List[List[str]]
        """
        return self.composers

    @COMPOSER.setter
    def COMPOSER(self, value: NSList):
        self.composers = value

    @property
    def COVERART(self) -> bytearray:
        """ Holds coverart read into memory as a bytes object. 

        :type: bytearray
        """
        return self.cover_art

    @COVERART.setter
    def COVERART(self, value: bytearray):
        self.cover_art = value

    @property
    def DATE(self) -> str:
        """ Album release date.

        :type: str
        """
        return self.release_date

    @DATE.setter
    def DATE(self, value: str):
        self.release_date = value

    @property
    def DISCNUMBER(self) -> IList:
        """List containing dicsnumber for every track.

        :type: List[int]
        """
        return self.disc_num

    @DISCNUMBER.setter
    def DISCNUMBER(self, value: IList):
        self.disc_num = value

    @property
    def GENRE(self) -> str:
        """If :attr:`genres` is a list with one item, than it is that item,
        otherwise user input is required to select from genres list.

        :type: str
        """
        return self.selected_genre

    @GENRE.setter
    def GENRE(self, value: str):
        self.selected_genre = value

    @property
    def LYRICS(self) -> SList:
        """List of lyrics coresponding to each track.

        :type: List[str]
        """
        return self.lyrics

    @LYRICS.setter
    def LYRICS(self, value: SList):
        self.lyrics = value

    @property
    def TITLE(self) -> SList:
        """ List of track names.

        :type: List[str]
        """
        return self.tracks

    @TITLE.setter
    def TITLE(self, value: SList):
        self.tracks = value

    @property
    def TRACKNUMBER(self) -> SList:
        """ List of numbers for each track.

        :type: List[str]
        """
        return self.numbers

    @TRACKNUMBER.setter
    def TRACKNUMBER(self, value: SList):
        self.numbers = value

    @property
    def FILE(self) -> SList:
        """ List of files on local disk corresponding to each track.

        :type: List[str]
        """
        return self.files

    @FILE.setter
    def FILE(self, value: SList):
        self.files = value

    @property
    def TYPE(self) -> SList:
        """ List of track types.

        :type: List[str]
        """
        return self.bracketed_types

    @TYPE.setter
    def TYPE(self, value: SList):
        self._bracketed_types = value


# TODO needs QThreads
"""
from wiki_music.gui.qt_importer import Signal, QObject


class GuiInOut(ParserInOut, QObject):

    selectedGenreChanged = Signal(str)
    releaseDateChanged = Signal(str)
    bandChanged = Signal(str)
    albumChanged = Signal(str)

    def __init__(self):

        QObject.__init__(self)
        self._selected_genre = ""
        self._release_date = ""
        self._album = ""
        self._band = ""

    @property
    def selected_genre(self):
        return self._selected_genre

    @selected_genre.setter
    def selected_genre(self, value):
        self._selected_genre = value
        self.selectedGenreChanged.emit(value)

    @property
    def release_date(self):
        return self._release_date

    @release_date.setter
    def release_date(self, value):
        self._release_date = value
        self.releaseDateChanged.emit(value)

    @property
    def band(self):
        return self._band

    @band.setter
    def band(self, value):
        self._band = value
        self.selectedGenreChanged.emit(value)

    @property
    def album(self):
        return self._album

    @album.setter
    def album(self, value):
        self._album = value
        self.selectedGenreChanged.emit(value)
"""



"""Logger initialization for package."""

import logging
from os import makedirs, path
from typing import Set

from wiki_music.constants.paths import LOG_DIR

logging.getLogger(__name__)

__all__ = ["set_log_handles"]

# logger formater
FORMATTER = logging.Formatter("%(asctime)s - %(levelname)s \n\t - "
                              "pid = %(process)d \n\t - "
                              "proces name = %(processName)s \n\t - "
                              "module = %(module)s,"
                              "funcName = %(funcName)s \n\t - "
                              "%(message)s \n\t",
                              datefmt="%H:%M:%S")

# create dir to store logs
makedirs(LOG_DIR, exist_ok=True)


def _compile_log_path(name: str) -> str:
    """Compiles standard logger path consisting of wiki_music_{name}.log

    Parameters
    ----------
    name: str
        unique logger name

    Returns
    -------
    str
        compiled logger path name
    """

    return path.join(LOG_DIR, f"wiki_music_{name}.log")


def set_log_handles(level: int):
    """Set desired level for package loggers and add file handlers.

    Parameters
    ----------
    level: int
        logging level
    """

    already_set: Set[str] = set()
    for name in logging.root.manager.loggerDict:
        if "wiki_music" in name:
            # assumed format is wiki_music.<submodule>.<subsubmodule>.<...>
            try:
                log_name = name.split(".")[1]
            except IndexError:
                pass
            else:
                if log_name in already_set:
                    continue

                log_path = _compile_log_path(log_name)

                fh = logging.FileHandler(log_path, mode="w")
                fh.setLevel(level)
                fh.setFormatter(FORMATTER)

                log = logging.getLogger(name)
                log.setLevel(level)
                log.addHandler(fh)

                already_set.add(log_name)

    return log



""" The base module for Qt frontend. """

import ctypes
import logging
from abc import abstractmethod

from wiki_music.constants import MAIN_WINDOW_UI
from wiki_music.gui_lib.qt_importer import (QAbstractItemView, QIcon,
                                            QMainWindow, QMessageBox,
                                            QStandardItemModel,
                                            QSystemTrayIcon, uic)
from wiki_music.utilities import MultiLog, abstract_warning, get_icon

log = logging.getLogger(__name__)
log.debug("base imports done")


# inherit base from QMainWindow and lyaout from Ui_MainWindow
class BaseGui(QMainWindow):
    """Base class for all GUI classes, initializes UI from Qt Designer
    generated files. then sets up needed variables. Connects buttons and input
    fields signals to methods. All GUI classes should subclass this class.

    Warnings
    --------
    This class is not ment to be instantiated, only inherited.

    Attributes
    ----------
    work_dir: str
        points to actuall selected directory with music files
    log: :class:`wiki_music.utilities.utils.MultiLog`
        class logger
    """

    def __init__(self) -> None:

        log.debug("init base")

        # call QMainWindow __init__ method
        super().__init__()
        # call Ui_MainWindow user interface setup method
        uic.loadUi(MAIN_WINDOW_UI, self)

        # initialize
        self.__initUI__()

        # misc
        self.work_dir: str = ""
        self.log: MultiLog = MultiLog(log)

        log.debug("init base done")

    def __initUI__(self):
        """ Has three responsibilities: load and set window and tray icon and
        Set application name.
        """

        self.setWindowTitle("Wiki Music")
        myappid = "WikiMusic"
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)
        _icon = get_icon()
        self.setWindowIcon(QIcon(_icon))
        tray_icon = QSystemTrayIcon(QIcon(_icon))
        tray_icon.show()

    def _do_nothing(self):
        """ Developement convenience function, shows messagebox with a warning
        about functionality not being implemented yet.
        """

        log.warning("Not implemented yet")
        QMessageBox(QMessageBox.Warning,
                    "Info", "Not implemented yet!").exec_()

    @abstractmethod
    def _display_image(self, image=None):
        """ Will be reimpemented in :mod:`wiki_music.gui_lib.data_model` module
        """
        abstract_warning()



"""Module for handling mp3 tags."""

import logging
from ast import literal_eval
from collections import OrderedDict
from typing import Dict, Union

from mutagen.id3 import (APIC, COMM, ID3, TALB, TCOM, TCON, TDRC, TIT2, TPE1,
                         TPE2, TPOS, TRCK, USLT, ID3NoHeaderError, PictureType)

from .tag_base import TagBase

log = logging.getLogger(__name__)
log.debug("loading mp3 module")


class TagMp3(TagBase):
    """A low level implementation of tag handling for mp3 files. """
    __doc__ += TagBase.__doc__  # type: ignore

    _map_keys = OrderedDict([
        (TALB, "ALBUM"),
        (TPE2, "ALBUMARTIST"),
        (TPE1, "ARTIST"),
        (COMM, "COMMENT"),
        (TCOM, "COMPOSER"),
        (TDRC, "DATE"),
        (TPOS, "DISCNUMBER"),
        (TCON, "GENRE"),
        (USLT, "LYRICS"),
        (TIT2, "TITLE"),
        (TRCK, "TRACKNUMBER"),
        (APIC, "COVERART")]
    )

    def _open(self, filename: str):
        """Function reading mp3 file to mutagen.id3.ID3 class. """

        try:
            self._song = ID3(filename=filename)
        except ID3NoHeaderError:
            print("Cannot read Mp3 tags")

    def _find_variable_key(self, tag: str) -> str:
        """Sometimes the key in tags is appended with some string like
        USTL::eng this functions finds the specific key if supliedd with the
        common prefix.

        Returns
        -------
        str
            mutagen.ID3 tag name
        """

        tag = self.reverse_map[tag]

        for k in self._song.keys():
            if tag in k:
                return self._key2str(k)

        return ""

    @staticmethod
    def _key2str(key):
        """get from string like <class 'mutagen.id3.TCOM'> the name TCOM.

        Returns
        -------
        str
            string reptresentation of mutagen.ID3 class tag name
        """
        return str(key).rsplit(".", 1)[1][:-2]

    def _read(self) -> Dict[str, Union[str, bytearray]]:

        tags = dict()
        for key, value in self._map_keys.items():  # pylint: disable=no-member
            key = self._key2str(key)
            try:
                # lyrics key is formated like: USLT::xxx
                # where xxx designates language
                if value == "LYRICS":
                    key = self._find_variable_key(value)
                    # lyrics tag field returns literal string
                    tag = literal_eval(self._song[key].text)
                elif value == "COVERART":
                    key = self._find_variable_key(value)
                    tag = self._song[key].data
                    continue
                else:
                    tag = self._song[key].text

                if isinstance(tag, list):
                    tag = tag[0]
                tag = str(tag).strip()

            except KeyError:
                tag = ""
            finally:
                tags[value] = tag

        return tags

    def _write(self, tag: str, value: Union[str, bytearray]):

        if tag == "COVERART":
            self._song[self.reverse_map[tag]](
                mime=u"image/jpeg", type=PictureType.COVER_FRONT,
                desc=u"Cover", data=value, encoding=3)
        else:
            self._song[self.reverse_map[tag]](encoding=3, text=value)



"""Utility functions and classe for whole wiki_music."""

import logging

from .exceptions import *
from .gui_utils import *
from .loggers import *
from .parser_utils import *
from .sync import *
from .utils import *
from .wrappers import *

logging.getLogger(__name__)



"""Module containing Main GUI window class called by GUI entry point."""

import logging
import os
import subprocess  # lazy loaded
import sys
import time  # lazy loaded
import webbrowser  # lazy loaded
from threading import Thread
from typing import Optional, Union

from wiki_music.constants import API_KEY_MESSAGE, MP3_TAG, ROOT_DIR
from wiki_music.gui_lib import BaseGui, CoverArtSearch, DataModel, RememberDir
from wiki_music.gui_lib.qt_importer import (QAbstractItemView, QFileDialog,
                                            QInputDialog, QMessageBox, QTimer)
from wiki_music.utilities import (Mp3tagNotFoundException, SharedVars,
                                  exception, synchronized, warning,
                                  we_are_frozen)

log = logging.getLogger(__name__)
log.debug("finished gui imports")


class Checkers(BaseGui):
    """Class that houses methods that periodically check for parser changes.

    Warnings
    --------
    This class is not ment to be instantiated, only inherited.
    """

    def __init__(self) -> None:

        super().__init__()

        self._remember: Optional[str] = None

    def _init_checkers(self):
        """Initializes timers for periodically repeating methods.

        See also
        --------
        :meth:`Checkers._conditions_check`
            checks if parser requires user input
        :meth:`Checkers._exception_check`
            checks foe exceptions in application and displays a messagebox
            with errors
        :meth:`Checkers._conditions_check`
            checks for parser progres descritions
        :meth:`Checkers._start_checkers`
            starts the initialized timers
        """
        self.description_timer = QTimer()
        self.description_timer.timeout.connect(self._description_check)
        self.description_timer.setInterval(400)

        self.exception_timer = QTimer()
        self.exception_timer.timeout.connect(self._exception_check)
        self.exception_timer.setInterval(500)

        self.conditions_timer = QTimer()
        self.conditions_timer.timeout.connect(self._conditions_check)
        self.conditions_timer.setSingleShot(True)

    def _start_checkers(self):
        """Initializes timers of periodically repeating methods.

        See also
        --------
        :meth:`Checkers._init_checkers`
            initializes timers
        """

        self.description_timer.start(200)
        self.exception_timer.start(500)
        self.conditions_timer.start(400)

    # periodically checking methods
    @exception(log)
    def _conditions_check(self):
        """Provides GUI<->user interaction for the parser.

        Periodically checks if parser requires user input, if so displays
        dialog with appropriate description and input options.
        """

        def msg_process(out: QMessageBox.StandardButton) -> Union[str, bool]:
            if out == QMessageBox.Yes:
                return True
            elif out == QMessageBox.No:
                return False
            else:
                return "d"

        if SharedVars.wait:

            if SharedVars.switch == "genres":

                log.debug("initialize question window")
                # TODO non-atomic
                if not self.genres:
                    text, ok = QInputDialog.getText(self, "Input genre",
                                                    "Input genre:")
                    if ok:
                        self.GENRE = text
                    else:
                        self.GENRE = "N/A"
                else:
                    item, ok = QInputDialog.getItem(self, "Select genre",
                                                    "List of genres",
                                                    self.genres, 0, False)
                    if ok:
                        self.GENRE = item
                    else:
                        self.GENRE = "N/A"

            if SharedVars.load:
                QTimer.singleShot(0, self._parser_to_gui)

            if SharedVars.switch == "comp":
                log.debug("ask to copy composers")
                msg = QMessageBox(QMessageBox.Question, "Question",
                                  "Do you want to copy artists to composers?",
                                  QMessageBox.Yes | QMessageBox.No)

                SharedVars.assign_artists = msg_process(msg.exec_())

            if SharedVars.switch == "lyrics":
                log.debug("ask to find lyrics")
                msg = QMessageBox(QMessageBox.Question, "Question",
                                  "Do you want to find lyrics?",
                                  QMessageBox.Yes | QMessageBox.No)

                SharedVars.write_lyrics = msg_process(msg.exec_())

            if SharedVars.switch == "api_key":
                log.debug("ask to get google api key")
                msg = QMessageBox(QMessageBox.Information, "Warning",
                                  API_KEY_MESSAGE.replace("\n", ""),
                                  QMessageBox.Yes | QMessageBox.No | 
                                  QMessageBox.Ignore)
                # change button text
                dont_bother_button = msg.button(QMessageBox.Ignore)
                dont_bother_button.setText("Don't bother me again")

                SharedVars.get_api_key = msg_process(msg.exec_())

            if SharedVars.switch == "load_api_key":
                log.debug("ask to input google api key")
                dialog = QInputDialog(self)
                dialog.setInputMode(QInputDialog.TextInput)
                dialog.setLabelText("Key:")
                dialog.setWindowTitle("Paste goole API key")
                dialog.resize(350, 70)                          

                if dialog.exec_():
                    SharedVars.get_api_key = dialog.textValue()
                else:
                    SharedVars.get_api_key = None

            SharedVars.load = False
            SharedVars.switch = ""
            SharedVars.wait = False

        if not SharedVars.done:
            self.conditions_timer.start(200)
        else:
            # announce that gui has reached the barrier
            log.debug("gui reached barrier")
            # TODO barrier is a wrong thing to have in GUI
            SharedVars.barrier.wait()

            log.debug("start _parser_to_gui function")
            QTimer.singleShot(0, self._parser_to_gui)

    def _exception_check(self):
        """Provides GUI displaying of exceptions.

        Periodically checks if exception occured somewhere in the app, if so
        displays message with information and if applicable option to exit the
        app
        """

        if SharedVars.has_exception:
            msg = QMessageBox(QMessageBox.Critical, "Exception",
                              SharedVars.has_exception)
            # TODO set right logging file
            msg.setDetailedText(open("logger","r").read())
            msg.exec_()
            SharedVars.has_exception = ""

        if SharedVars.has_warning:
            QMessageBox(QMessageBox.Warning, "Warning",
                        SharedVars.has_warning).exec_()
            SharedVars.has_warning = ""

        if SharedVars.ask_exit:
            msg = QMessageBox(QMessageBox.Question, "Warning",
                              "Do you want to stop the search?",
                              QMessageBox.Yes | QMessageBox.No)
            msg.setInformativeText(SharedVars.ask_exit)
            terminate = msg.exec_()

            if terminate == QMessageBox.Yes:
                SharedVars.wait_exit = False
                SharedVars.terminate_app = True

                time.sleep(0.03)
                # TODO not working
                sys.exit(self.exec_())  # TODO untested
            else:
                SharedVars.wait_exit = False
                SharedVars.ask_exit = ""

    # TODO this method should also update progessbar in status bar
    # lock must be here because there are non atomic operations i.e. +=
    @exception(log)
    @synchronized(SharedVars.lock)
    def _description_check(self):
        """Shows parser progress in the bottom status bar of main window.

        Runs periodically.
        """

        if " . . ." in SharedVars.describe:
            SharedVars.describe = SharedVars.describe.replace(" . . .", "")

        self._remember = SharedVars.describe

        if SharedVars.describe.strip():
            if (self._remember == SharedVars.describe and
                "Done" not in SharedVars.describe):  # noqa E129
                SharedVars.describe += " ."
            self.statusbar.showMessage(SharedVars.describe)
        else:
            self.statusbar.showMessage("")


class Buttons(BaseGui):
    """Class for handling main window button interactions.

    Warnings
    --------
    This class is not ment to be instantiated, only inherited.
    """

    @exception(log)
    def __open_dir__(self):
        """Opens the current search folder."""

        if self.work_dir:
            self.log.info("Opening folder...")
            os.startfile(self.work_dir)
        else:
            QMessageBox(QMessageBox.Information, "Message",
                        "You must select directory first!").exec_()

    def _select_file(self, description: str = "Select song file",
                     file_types: str = "Audio files (*.mp3 *.flac *.m4a)"
                     ) -> str:
        """Handle file selection
        
        Parameters
        ----------
        description: str
            window title
        file_types: str
            string with allowed types

        Returns
        -------
        str
            path to selected file
        """
        _file = QFileDialog.getOpenFileName(self, description, self.work_dir,
                                            f"{file_types};;All files (*)")

        self.file_detail.setText(_file[0])
        return _file[0]


    @exception(log)
    def _open_browser(self, url: Optional[str] = None):
        """Opens the found wikipedia page in default web browser.

        Parameters
        ----------
        url: Optional[str]
            force opening of input url instead of parser wikipedia url
        """

        if url:
            webbrowser.open_new_tab(url)
        else:
            if self.url and not SharedVars.offline_debbug:
                webbrowser.open_new_tab(self.url)
            elif SharedVars.offline_debbug:
                self.log.warning("You are in offline debugging mode")
            else:
                QMessageBox(QMessageBox.Information, "Message",
                            "You must run the search first!").exec_()

    @exception(log)
    @warning(log)
    def _run_Mp3tag(self):
        """Open Mp3tag in current working directory.

        Warnings
        --------
        This action works only in Windods if Mp3tag is not installed in the
        right directory a dialog to set the path is displayed.
        """

        global MP3_TAG
        path_file = os.path.join(ROOT_DIR, "files", "MP3_TAG_PATH.txt")

        if os.path.isfile(path_file):
            with open(path_file, "r") as f:
                MP3_TAG = f.read().strip()

        if not MP3_TAG:
            msg = QMessageBox(QMessageBox.Warning, "Unable to locate Mp3tag",
                              "Do you want to set the location?",
                              QMessageBox.Yes | QMessageBox.No)
            if msg.exec_() == QMessageBox.No:
                return

        if sys.platform == "win32":
            if self.work_dir:
                try:
                    subprocess.Popen([MP3_TAG, f"/fp:{self.work_dir}"])
                except (FileNotFoundError, OSError):
                    # get loaction from user
                    MP3_TAG = self._select_file(
                        description="Mp3tag was not found, Select Mp3tag.exe "
                                    "executable path",
                        file_types="Executable files (*.exe)")

                    # write to file for later uses
                    with open(path_file, "w") as f:
                        f.write(MP3_TAG)
                    # run again
                    self._run_Mp3tag()
            else:
                QMessageBox(QMessageBox.Information, "Message",
                            "You must select directory first!").exec_()
        else:
            raise Mp3tagNotFoundException("Mp3tag is supported only on "
                                          "Windows")

    def _show_help(self, help_type):

        if help_type == "idx":
            self._open_browser("https://wikipedia-music-tags.readthedocs.io"
                               "/en/latest/index.html")
        elif help_type == "about":
            self._open_browser("https://github.com/marian-code/"
                               "wikipedia-music-tags")
        elif help_type == "git":
            self._open_browser("https://github.com/marian-code/"
                               "wikipedia-music-tags")

    def __entry_band__(self):
        """Connect to albumartist entry field."""
        self.ALBUMARTIST = self.band_entry_input.text()
        self.start_preload()

    def __entry_album__(self):
        """Connect to album entry field."""
        self.ALBUM = self.album_entry_input.text()
        self.start_preload()

    def __select_json__(self):
        """Connect to json checkbox."""
        SharedVars.write_json = self.json_write_sw.isChecked()

    def __select_offline_debbug__(self):
        """Connect to offline debug checkbox.
        
        Note
        ----
        If the checkbox is unselected preload is stopped.
        """
        SharedVars.offline_debbug = self.offline_debbug_sw.isChecked()
        if SharedVars.offline_debbug:
            self.stop_preload()


class Window(DataModel, Checkers, Buttons, CoverArtSearch):
    """Toplevel GUI class, main winndow with all its functionality."""

    def __init__(self, debug):

        log.debug("init superclass")
        super().__init__()

        # whether to show debuggingoptions in gui
        self._DEBUG = debug

        log.debug("setup overlay")
        # set overlay functions
        self.__setup_overlay__()

        log.debug("start checkers")
        # start checkers
        self._init_checkers()
        self._start_checkers()

    # setup methods
    def __setup_overlay__(self):
        """Sets up GUI input elements siganl connections."""

        # map buttons to functions
        # must use lambda otherwise wrapper doesnt work correctly
        self.browse_button.clicked.connect(lambda: self.__select_dir__())
        self.wiki_search_button.clicked.connect(lambda: self.__run_search__())
        self.coverArt.clicked.connect(lambda: self.cover_art_search())
        self.lyrics_button.clicked.connect(
            lambda: self.__run_lyrics_search__())
        self.toolButton.clicked.connect(lambda: self._select_file())

        # connect the album and band entry field to preload function
        self.band_entry_input.editingFinished.connect(self.__entry_band__)
        self.album_entry_input.editingFinished.connect(self.__entry_album__)

        # create table-proxy mapping for sorting
        self.tableView.setModel(self.proxy)
        self.tableView.setSortingEnabled(True)

        # show table
        self.tableView.show()

        # enable drag and drop ordering of columns in table
        self.tableView.horizontalHeader().setSectionsMovable(True)
        self.tableView.horizontalHeader().setDragEnabled(True)
        self.tableView.horizontalHeader().setDragDropMode(
            QAbstractItemView.InternalMove)

        # connect to signal that is emited when table cell is clicked
        self.tableView.clicked.connect(self._detail)

        # connect menubar actions to functions
        # must use lambda otherwise wrapper doesnt work correctly
        self.actionDirectory.triggered.connect(lambda: self.__open_dir__())
        self.actionWikipedia.triggered.connect(lambda: self._open_browser())
        self.actionMp3_Tag.triggered.connect(lambda: self._run_Mp3tag())
        self.actionAll_Tags.triggered.connect(lambda: self.__save_all__())
        # TODO get rid of save only lyrics button, we now save inteligently
        # TODO only changed tags
        self.actionOnly_Lyrics.triggered.connect(lambda: self.__save_all__())

        # show help
        self.actionHelp_Index.triggered.connect(lambda: self._show_help("idx"))
        self.actionAbout.triggered.connect(lambda: self._show_help("about"))
        self.actionGit.triggered.connect(lambda: self._show_help("git"))

        # TODO menubar buttons taht are not implemented
        self.actionNew.triggered.connect(self._do_nothing)
        self.actionOpen.triggered.connect(self._do_nothing)
        self.actionExit.triggered.connect(self._do_nothing)
        self.actionSave.triggered.connect(self._do_nothing)

        # TODO this needs to be hidden alfo for pip
        if not self._DEBUG:
            # show switches only if not frozen
            self.offline_debbug_sw.hide()
            self.json_write_sw.hide()
        else:
            # connect switches to functions
            self.offline_debbug_sw.stateChanged.connect(
                self.__select_offline_debbug__)
            self.json_write_sw.stateChanged.connect(self.__select_json__)

    # methods that bind to gui elements
    @exception(log)
    def __save_all__(self):
        """Save changes to file tags, after saving reload files from disk."""

        # first stop any running preload, as it is not needed any more
        self.stop_preload()

        # TODO non-atomic
        if not self.save_ready:
            QMessageBox(QMessageBox.Information, "Info",
                        "You must run the search first!").exec_()
        else:
            if self.work_dir == "":
                QMessageBox(QMessageBox.Information, "Info",
                            "You must specify directory with files!").exec_()
                return

            if not self.write_tags():
                msg = ("Cannot write tags because there are no "
                       "coresponding files")
                QMessageBox(QMessageBox.Information, "Info", msg)
                self.log.info(msg)

            # reload files from disc after save
            self.reinit_parser()
            self.read_files()
            self._parser_to_gui()

    @exception(log)
    def __select_dir__(self):
        """Select working directory, read found files and start preload."""

        with RememberDir(self) as rd:
            self.work_dir = rd.get_dir()

        # TODO non-atomic
        # read files and start preload
        self.read_files()
        self.start_preload()

        self._parser_to_gui()

    def __check_input_is_present__(self) -> bool:
        """Check if apropriate input to conduct search is present.

        Returns
        -------
        bool
            true if all requested inputs are present
        """

        group = [self.ALBUMARTIST, self.ALBUM, self.work_dir]
        if all(group):
            return True
        else:
            alb, bnd, wkd, inp, com, _and = [""] * 6
            if not self.ALBUM:
                alb = " album"
            if not self.ALBUMARTIST:
                bnd = " band"
            if not self.work_dir:
                wkd = " select working directory"

            if any([alb, bnd]):
                inp = "input"
            if all([alb, bnd, wkd]):
                com = ","
            elif all([alb, bnd]):
                com = " and"
            if all([alb, wkd]) or all([wkd, bnd]):
                _and = " and"

            msg = f"You must {inp}{alb}{com}{bnd}{_and}{wkd}"

            message = QMessageBox()
            message.setIcon(QMessageBox.Information)
            message.setWindowTitle("Message")
            message.setText(msg)
            message.exec_()
            self.log.warning(msg)

            return False

    @exception(log)
    def __run_search__(self):
        """Start wikipedia search in background thread."""

        if not self.__check_input_is_present__():
            return

        log.info("starting wikipedia search")

        # TODO non-atomic
        self.reinit_parser()
        self._start_checkers()

        main_app = Thread(target=self._parser.run_wiki, name="WikiSearch")
        main_app.daemon = True
        main_app.start()

    @exception(log)
    def __run_lyrics_search__(self):
        """Start only lyric search in background thread."""

        self.stop_preload()

        if not self.__check_input_is_present__():
            return

        log.info("starting lyrics search")

        SharedVars.write_lyrics = True

        self._start_checkers()
        main_app = Thread(target=self._parser.run_lyrics, name="LyricsSearch")
        main_app.daemon = True
        main_app.start()



""" Module for importing all Qt classes used by GUI.

Note
----
With the help of QtPy all major Qt bindings are supported: PyQt5, PyQt4,
Pyside2 and PySide.
"""

# debug readthedocs build

try:
    from PyQt5.QtCore import PYQT_VERSION_STR
except ImportError as e:
    print(e)
try:
    from PyQt5.QtCore import QT_VERSION_STR
except ImportError as e:
    print(e)
try:
    from PySide2 import __version__
except ImportError as e:
    print(e)
try:
    from PySide2.QtCore import __version__
except ImportError as e:
    print(e)
# end of reradthedocs debug

import logging

log = logging.getLogger(__name__)

try:
    from qtpy.QtWidgets import (QMainWindow, QFileDialog, QApplication,
                                QTableWidgetItem, QMessageBox, QHBoxLayout,
                                QAbstractItemView, QInputDialog, QLabel,
                                QVBoxLayout, QTableWidget, QWidget, QDialog,
                                QStatusBar, QSystemTrayIcon, QSizePolicy,
                                QRubberBand, QStyleFactory, QSizeGrip)
    from qtpy.QtGui import (QStandardItemModel, QStandardItem, QImage,
                            QPixmap, QIcon, QPainter, QResizeEvent)
    from qtpy.QtCore import (Qt, QSortFilterProxyModel, QTimer, QObject,
                             QBuffer, QSize, QRect, Property, Slot, Signal,
                             QByteArray, QIODevice, QPoint, QEvent,
                             QModelIndex, QVariant)
    from qtpy import uic
except ImportError as e:
    log.critical("None of the Qt backends is available! Aborting")
    raise ImportError("None of the Qt backends is available! Aborting")
else:
    log.debug("Qt imports done")



"""This module houses custom Qt classes designed to support the package."""

import logging
from os import makedirs, path
from typing import Callable, Iterable, Optional, Tuple, Union

from wiki_music.constants.paths import DIR_FILE
from wiki_music.gui_lib.qt_importer import (Property, QBuffer, QByteArray,
                                            QEvent, QFileDialog, QHBoxLayout,
                                            QImage, QIODevice, QLabel,
                                            QModelIndex, QPixmap, QPoint,
                                            QRect, QResizeEvent, QRubberBand,
                                            QSize, QSizeGrip, QSizePolicy,
                                            QSortFilterProxyModel,
                                            QStandardItem, QStandardItemModel,
                                            QStyleFactory, Qt, QTableWidget,
                                            QVariant, QVBoxLayout, QWidget,
                                            Signal)
from wiki_music.utilities.gui_utils import get_music_path

logging.getLogger(__name__)

__all__ = ["NumberSortModel", "CustomQStandardItem", "ImageTable",
           "ResizablePixmap", "CustomQStandardItemModel", "RememberDir"]


class NumberSortModel(QSortFilterProxyModel):
    """Custom table proxy model which can sort numbers not only strings."""

    def lessThan(self, left_index: QModelIndex,
                 right_index: QModelIndex) -> bool:
        """Reimplemented comparator method to handle numbers correctly.

        Parameters
        ----------
        left_index: QModelIndex
            index of the left compared element
        right_index: QModelIndex
            index of the right compared element
        """

        left_var: str = left_index.data(Qt.EditRole)
        right_var: str = right_index.data(Qt.EditRole)

        try:
            return float(left_var) < float(right_var)
        except (ValueError, TypeError):
            pass

        try:
            return left_var < right_var
        except TypeError:  # in case of NoneType
            return True


class CustomQStandardItem(QStandardItem):
    """Overrides default methods so only the filename is "displayed"
    even though the object stores the full path.

    Parameters
    ----------
    data: Union[str, int, float, Iterable[str]]
        supports all data types used by parser, all are converted to string
        before passing to superclass constructor
    """

    def __init__(self, data: Union[str, int, float, Iterable[str]]) -> None:

        if isinstance(data, str):
            pass
        elif isinstance(data, (int, float)):
            data = str(data)
        elif isinstance(data, (list, tuple)):
            data = ", ".join(sorted(data))
        else:
            data = ""

        super().__init__(data)
        self._filtered: str = ""

    def data(self, role: Qt.ItemDataRole) -> QVariant:
        """Reimplemented method to show only filename instead of full file
        path. Only Qt.DisplayRole is affected other roles are left untouched.

        Parameters
        ----------
        role: Qt.ItemDataRole
            role of the edited item

        Returns
        -------
        Any
            data contained in QStandardItem
        """

        if role != Qt.DisplayRole:
            return super().data(role)
        else:
            if not self._filtered:

                self._filtered = super().data(Qt.DisplayRole)
                if path.isfile(self._filtered):
                    self._filtered = path.basename(self._filtered)

            return self._filtered

    def setData(self, value: QVariant, role: Qt.UserRole):
        """ Reimplemented, sets new data for QStandardItem.
        also sets the :attr:`_filtered` to None so the path for the new item
        can be filtered again.

        Parameters
        ----------
        value: QVariant
            new value to set
        role: Qt.UserRole
            Qt data role
        """
        print("getting")
        self._filtered = ""
        super().setData(value, role)

    def real_data(self, role: Qt.ItemDataRole) -> str:
        """ Workaround to show real contained data because the
        :meth:`QStandardItem.data` method is overridden.

        Parameters
        ----------
        role: Qt.ItemDataRole
            role of the edited item

        Returns
        -------
        Any
            data contained in QStandardItem
        """

        return super().data(role)

    def text(self, split=False) -> Union[str, list]:
        """Reimplemented, if the data is path than return full path instead
        of only filename.

        Parameters
        ----------
        split: bool
            if true split string into list of strings on commas

        Returns
        -------
        Union[str, list]
            string or list with actual data based on input
        """
        text = super().text()
        data = self.real_data(Qt.DisplayRole)

        if len(data) < len(text):
            data = text

        if split and "," in data:
            return [x.strip() for x in data.split(",")]


class CustomQStandardItemModel(QStandardItemModel):
    """ Overrides the default impementation adds `__getitem__` method so the
    table columns ca be indexed by its names.
    """

    def __getitem__(self, name: str) -> int:
        """ Column index from column header name.

        Parameters
        ----------
        name: str
            table column name

        Returns
        -------
        int
            column index

        Raises
        ------
        KeyError
            if the name does not exist in column headers
        """

        for column in range(self.columnCount()):
            if self.headerData(column, Qt.Horizontal) == name:
                return column

        raise KeyError(f"No column with name {name}")


class ImageWidget(QWidget):
    """Widget that displays image along with a short string underneath.

    Attributes
    ----------
    text: str
        test to show under picture
    img: bytearray
        picture to show
    """

    def __init__(self, text: str, img: bytearray,
                 parent: Optional[QWidget] = None) -> None:
        QWidget.__init__(self, parent)

        self._text = text
        self._img = img

        self.setLayout(QVBoxLayout())
        self.lbPixmap = QLabel(self)
        self.lbText = QLabel(self)
        self.lbText.setAlignment(Qt.AlignCenter)

        self.layout().addWidget(self.lbPixmap)
        self.layout().addWidget(self.lbText)

        self.initUi()

    def initUi(self):
        """ Sets the text and image to be visible. """

        image = QImage()
        image.loadFromData(self._img)
        self.lbPixmap.setPixmap(QPixmap(image).scaledToWidth(150))
        self.lbText.setText(self._text)

    @Property(bytearray)
    def img(self) -> bytearray:
        """Dispalyed image in bytes format.

        :type: bytearray
        """
        return self._img

    @img.setter
    def ing_setter(self, value: bytearray):
        if self._img == value:
            return
        self._img = value
        self.initUi()

    @Property(str)
    def text(self) -> str:
        """Short text under the picture.

        :type: str
        """
        return self._text

    @text.setter
    def text_setter(self, value: str):
        if self._text == value:
            return
        self._text = value
        self.initUi()


class ImageTable(QTableWidget):
    """Table with cells holding Images with short text in cells.

    Parameters
    ----------
    parent: QWidget
        reterence to parent widget for centering

    Attribures
    ----------
    self.MAX_COLUMNS: int
        constant saying how many columns should the table have
    self.actualCol: int
        holds the position of actual column

    See also
    --------
    :class:ImageWidget
        each cell holds one of these widgets
    """

    MAX_COLUMNS: int
    actualCol: int

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        QTableWidget.__init__(self, parent)

        self.MAX_COLUMNS: int = 4
        self.actualCol: int = -1
        self.setColumnCount(1)
        self.setRowCount(1)
        self.resizeToContents()
        self.horizontalHeader().setVisible(False)
        self.verticalHeader().setVisible(False)
        self.setShowGrid(False)

    def add_pic(self, label: str, picture: bytearray):
        """Add one picture to table.

        Pictures are added in row until the row has `ImageTable.MAX_COLUMNS`
        pictures then we switch to the next row.

        Parameters
        ----------
        label: str
            short text to display under picture
        picture: bytearray
            picture to display

        Raises
        ------
        ValueError
            when method trie to add picture in column with
            `index` > `ImageTable.MAX_COLUMNS`
        """

        if self.actualCol < self.MAX_COLUMNS:
            self.actualCol += 1
            if self.rowCount() == 1:
                self.setColumnCount(self.columnCount() + 1)
        elif self.actualCol == self.MAX_COLUMNS:
            self.actualCol = 0
            self.setRowCount(self.rowCount() + 1)
        else:
            raise ValueError("Exception in adding picture to table. Actual"
                             " column exceeded value of max columns")

        self.setCellWidget(*self.ij, ImageWidget(label, picture))
        self.resizeToContents()

    @property
    def ij(self) -> Tuple[int, int]:
        """Actual free cell where next picture should be added.

        :type:`Tuple[int, int]`
        """
        return self.rowCount() - 1, self.actualCol

    def resizeToContents(self):
        """Resizes table rows and columns to fit largest cell contents."""
        self.resizeColumnsToContents()
        self.resizeRowsToContents()


class ResizableRubberBand(QRubberBand):
    """Reimplements QRubberBand so its aspect ratio can be set.

    Atttributes
    -----------
    aspect_ratio: Tuple[int, int]
        the aspect ratio that the rubberband is forcet to obey
    """

    aspect_ratio: Optional[float]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.aspect_ratio = None

    def resizeEvent(self, event: QEvent):
        """Reimplements resizeEvent with forced aspect ratio.

        Parameters
        ----------
        event: QEvent
            the event that triggered this method
        """

        if self.aspect_ratio:
            size = QSize(*self.aspect_ratio)
            size.scale(self.size(), Qt.KeepAspectRatio)
            self.resize(size)

    def set_aspect_ratio(self, ratio: float):
        """Sets forced aspect ratio for rubberband

        Parameters
        ----------
        ratio: float
            aspect ratio value, one float number
        """
        self.aspect_ratio = ratio
        self.resizeEvent(None)


class ResizablePixmap(QLabel):
    """Picture that can be arbitrarilly resized while keeping its aspect ratio.

    Parameters
    ----------
    bytes_image_edit: bytearray
        image that is displayed
    """

    bytes_image_edit: bytearray

    def __init__(self, bytes_image: bytearray) -> None:
        QLabel.__init__(self)
        self.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Ignored)
        self.setAlignment(Qt.AlignCenter)
        self.setStyleSheet("background-color: #ffffff;")

        self.update_pixmap(bytes_image)

    def update_pixmap(self, bytes_image: bytearray):
        """Changes displayed image for a new one.

        Parameters
        ----------
        bytes_image: bytearray
            new image to display in bytes format
        """
        self.bytes_image_edit = bytes_image
        self.current_pixmap = self._bytes2pixmap(bytes_image)
        self.scale()

    def scale(self, fromResize: bool = False):
        """Handle picture scaling.

        Parameters
        ----------
        fromResize: bool
            special handling if this method was called after resize, but only 
            in subclasses
        """

        # use a single central method for scaling; there's no need to call it
        # upon creation and also resize() won't work anyway in a layout
        self.setPixmap(self.current_pixmap.scaled(self.width(), self.height(),
                                                  Qt.KeepAspectRatio,
                                                  Qt.SmoothTransformation))

    def resizeEvent(self, event: QEvent):
        """Reimplement behaviour on resize with scaling enabled.

        Parameters
        ----------
        event: QEvent
            the event that triggered this method
        """

        super(ResizablePixmap, self).resizeEvent(event)
        self.scale(fromResize=True)

    @staticmethod
    def _bytes2pixmap(raw_image: bytearray) -> QPixmap:
        """Convert bytes image to `QPixmap`

        Parameters
        ----------
        raw_image: bytearray
            bytes image to be converted
        
        Returns
        -------
        QPixmap
            image as Qt QPixmap
        """

        image = QImage()
        image.loadFromData(raw_image)
        return QPixmap(image)

    @staticmethod
    def _pixmap2bytes(pixmap: QPixmap) -> bytearray:
        """ Convert `QPixmap` to bytes image.

        Parameters
        ----------
        pixmap: QPixmap
            Qt QPixmap to be converted
        
        Returns
        -------
        bytearray
            bytes image
        """

        byte_array = QByteArray()
        buffer = QBuffer(byte_array)
        buffer.open(QIODevice.WriteOnly)
        pixmap.save(buffer, 'PNG')
        return byte_array.data()

    @property
    def image_dims(self) -> Tuple[int, int]:
        """Actual image dimensions

        Returns
        -------
        Tuple[int, int]
            image width and height
        """
        return self.width(), self.height()


class SelectablePixmap(ResizablePixmap):
    """Pixmap whose part can be selected with selection rubberband.

    Warnings
    --------
    The solution is copied from stackoverflow and is not yet
    properly documented.

    References
    ----------
    https://stackoverflow.com/questions/58053735/get-real-size-of-qpixmap-in-qlabel

    Attributes
    ----------
    currentQRubberBand: Optional[ResizableRubberBand]
        holds reference to current rubberband
    rubberBandOffset: Optional[QPoint]
        rubberband upper left corner offset from the point where the mouse was
        clicked before dragging start
    moveDirection: Optional[Qt.Orientation]
        direction in which rubberband edge is dragged when resizing
    rubberBandRatio: Optional[float]
        forced aspect ratio of rubberband
    pixmapRect: QRect
        rectangle defining the pixmap borders
    selectionActive: Signal(bool)
        signal emited whenerver new rubberband selection is created
        destroyed
    """

    currentQRubberBand: Optional[ResizableRubberBand]
    rubberBandOffset: Optional[QPoint]
    moveDirection: Optional[Qt.Orientation]
    rubberBandRatio: Optional[float]
    pixmapRect: QRect

    selectionActive: Signal = Signal(bool)

    def __init__(self, bytes_image: bytearray) -> None:
        super().__init__(bytes_image)

        # activate mouse tracking to change cursor on rubberband hover
        self.setMouseTracking(True)
        self.currentQRubberBand = None
        self.rubberBandOffset = None
        self.moveDirection = 0
        self.rubberBandRatio = None

    def set_aspect_ratio(self, ratio: float):
        """Sets aspect ratio for rubberband.

        Parameters
        ----------
        ratio: float
            desired aspect ratio
        """
        self.rubberBandRatio = ratio
        self._update_ratio()

    def _update_ratio(self):
        """If rubberband is created updates its aspect ratio."""
        if self.currentQRubberBand:
            self.currentQRubberBand.set_aspect_ratio(self.rubberBandRatio)

    def create_selection(self, pos: QPoint):
        """Creates new rubberband selection.

        Parameters
        ----------
        pos: QPoint
            clicked position

        Note
        ----
        If old selection existed new is created only if mouse click happens
        outside of that selection. Othervise the current selection is moved.
        """

        if self.currentQRubberBand:
            self.cancel_selection()
        self.currentQRubberBand = ResizableRubberBand(QRubberBand.Rectangle,
                                                      self)
        self.currentQRubberBand.setStyle(QStyleFactory.create("Fusion"))
        self.currentQRubberBand.setGeometry(pos.x(), pos.y(), 1, 1)
        self.currentQRubberBand.show()
        self.originQPoint = pos
        self.currentQRubberBand.installEventFilter(self)

    def cancel_selection(self):
        """Cancels the current selection and destroys the rubberband."""

        self.currentQRubberBand.hide()
        self.currentQRubberBand.deleteLater()
        self.currentQRubberBand = None
        self.originQPoint = None
        self.selectionActive.emit(False)

    def scale(self, fromResize: bool = False):
        """ Handle picture and selection scaling caused by window resize.

        Parameters
        ----------
        fromResize: bool
            tells the type if event that requested scaling
        """

        if fromResize and self.currentQRubberBand:
            # keep data for rubber resizing, before scaling
            oldPixmapRect = self.pixmap().rect()
            oldOrigin = (self.currentQRubberBand.pos() -
                         self.pixmapRect.topLeft())
        super(SelectablePixmap, self).scale()

        # assuming that you always align the image in the center,
        # get the current pixmap rect and move
        # the rectangle center to the current geometry
        self.pixmapRect = self.pixmap().rect()
        self.pixmapRect.moveCenter(self.rect().center())
        if fromResize and self.currentQRubberBand:
            # find the new size ratio based on the previous
            xRatio = self.pixmapRect.width() / oldPixmapRect.width()
            yRatio = self.pixmapRect.height() / oldPixmapRect.height()
            # create a new geometry using 0-rounding for improved accuracy
            self.currentQRubberBand.setGeometry(
                round(oldOrigin.x() * xRatio, 0) + self.pixmapRect.x(),
                round(oldOrigin.y() * yRatio + self.pixmapRect.y(), 0),
                round(self.currentQRubberBand.width() * xRatio, 0),
                round(self.currentQRubberBand.height() * yRatio, 0))

    def updateMargins(self):
        """Update picture margins formouse event tracking.

        Whenever the rubberband rectangle geometry changes, create virtual
        rectangles for corners and sides to ease up mouse event checking.
        """
        rect = self.currentQRubberBand.geometry()
        self.rubberTopLeft = QRect(rect.topLeft(),
                                   QSize(8, 8))
        self.rubberTopRight = QRect(rect.topRight(),
                                    QSize(-8, 8)).normalized()
        self.rubberBottomRight = QRect(rect.bottomRight(),
                                       QSize(-8, -8)).normalized()
        self.rubberBottomLeft = QRect(rect.bottomLeft(),
                                      QSize(8, -8)).normalized()
        self.rubberLeft = QRect(self.rubberTopLeft.bottomLeft(),
                                self.rubberBottomLeft.topRight())
        self.rubberTop = QRect(self.rubberTopLeft.topRight(),
                               self.rubberTopRight.bottomLeft())
        self.rubberRight = QRect(self.rubberTopRight.bottomLeft(),
                                 self.rubberBottomRight.topRight())
        self.rubberBottom = QRect(self.rubberBottomLeft.topRight(),
                                  self.rubberBottomRight.bottomLeft())
        self.rubberInnerRect = QRect(self.rubberTop.bottomLeft(),
                                     self.rubberBottom.topRight())

    def eventFilter(self, source, event: QEvent):
        """Filteres GUI events to call special method on resize event.

        Parameters
        ----------
        source
            source that caused the event
        event: QEvent
            type of event
        
        Returns
        -------
        Callable
            result of the superclass eventFilter
        """

        if event.type() in (QEvent.Resize, QEvent.Move):
            self.updateMargins()
        return super(SelectablePixmap, self).eventFilter(source, event)

    def mousePressEvent(self, event: QEvent):
        """ Handles left mouse button cliks.

        If the clicked position is inside the current selection than that
        selection is moved. If it is outside tahn a new selection is created.
        If the click is in selection margins than the picture is resized by
        dragging one of the edges.

        Parameters
        ----------
        event: QEvent
            calling event

        See also
        --------
        :meth:`SelectablePixmap.updateMargins`
            method responsible for creating margins and keeping them up to date
        """

        pos = event.pos()

        if (not self.currentQRubberBand or
            pos not in self.currentQRubberBand.geometry()):
            if pos not in self.pixmapRect:
                self.originQPoint = None
                return
            self.create_selection(pos)
            self.selectionActive.emit(True)
            self._update_ratio()
        elif pos in self.rubberTopLeft:
            self.originQPoint = (
                self.currentQRubberBand.geometry().bottomRight())
        elif pos in self.rubberTopRight:
            self.originQPoint = self.currentQRubberBand.geometry().bottomLeft()
        elif pos in self.rubberBottomRight:
            self.originQPoint = self.currentQRubberBand.geometry().topLeft()
        elif pos in self.rubberBottomLeft:
            self.originQPoint = self.currentQRubberBand.geometry().topRight()
        elif pos in self.rubberTop:
            self.originQPoint = self.currentQRubberBand.geometry().bottomLeft()
            self.moveDirection = Qt.Vertical
        elif pos in self.rubberBottom:
            self.originQPoint = self.currentQRubberBand.geometry().topLeft()
            self.moveDirection = Qt.Vertical
        elif pos in self.rubberLeft:
            self.originQPoint = self.currentQRubberBand.geometry().topRight()
            self.moveDirection = Qt.Horizontal
        elif pos in self.rubberRight:
            self.originQPoint = self.currentQRubberBand.geometry().topLeft()
            self.moveDirection = Qt.Horizontal
        else:
            self.rubberBandOffset = pos - self.currentQRubberBand.pos()

    def mouseMoveEvent(self, event: QEvent):
        """Handles mouse movement events concerning the rubberband.

        The movement after mouse button is clicked and held can cause
        two actions. If the click occured in the margins than the selection is
        resized by dragging the edge. If it is inside the selection rectangle
        then the selection rubberband is moved by drgging.

        Parameters
        ----------
        event: QEvent
            the calling event
        """

        pos = event.pos()
        if event.buttons() == Qt.NoButton and self.currentQRubberBand:
            if pos in self.rubberTopLeft or pos in self.rubberBottomRight:
                self.setCursor(Qt.SizeFDiagCursor)
            elif pos in self.rubberTopRight or pos in self.rubberBottomLeft:
                self.setCursor(Qt.SizeBDiagCursor)
            elif pos in self.rubberLeft or pos in self.rubberRight:
                self.setCursor(Qt.SizeHorCursor)
            elif pos in self.rubberTop or pos in self.rubberBottom:
                self.setCursor(Qt.SizeVerCursor)
            elif pos in self.rubberInnerRect:
                self.setCursor(Qt.SizeAllCursor)
            else:
                self.unsetCursor()
        elif event.buttons():
            if self.rubberBandOffset:
                target = pos - self.rubberBandOffset
                rect = QRect(target, self.currentQRubberBand.size())
                # limit positioning of the selection to the image rectangle
                if rect.x() < self.pixmapRect.x():
                    rect.moveLeft(self.pixmapRect.x())
                elif rect.right() > self.pixmapRect.right():
                    rect.moveRight(self.pixmapRect.right())
                if rect.y() < self.pixmapRect.y():
                    rect.moveTop(self.pixmapRect.y())
                elif rect.bottom() > self.pixmapRect.bottom():
                    rect.moveBottom(self.pixmapRect.bottom())
                self.currentQRubberBand.setGeometry(rect)
            elif self.originQPoint:
                if self.moveDirection == Qt.Vertical:
                    # keep the X fixed to the current right, so that only the
                    # vertical position is changed
                    pos.setX(self.currentQRubberBand.geometry().right())
                else:
                    # limit the X to the pixmapRect extent
                    if pos.x() < self.pixmapRect.x():
                        pos.setX(self.pixmapRect.x())
                    elif pos.x() > self.pixmapRect.right():
                        pos.setX(self.pixmapRect.right())
                if self.moveDirection == Qt.Horizontal:
                    # same as before, but for the Y position
                    pos.setY(self.currentQRubberBand.geometry().bottom())
                else:
                    # limit the Y to the pixmapRect extent
                    if pos.y() < self.pixmapRect.y():
                        pos.setY(self.pixmapRect.y())
                    elif pos.y() > self.pixmapRect.bottom():
                        pos.setY(self.pixmapRect.bottom())
                rect = QRect(self.originQPoint, pos)
                self.currentQRubberBand.setGeometry(rect.normalized())

    def mouseReleaseEvent(self, event: QEvent):
        """Handles mouse release events and cleans up the data afterwards.
        
        Resets: rubberBandOffset, originQPoint, moveDirection
        """

        self.rubberBandOffset = None
        self.originQPoint = None
        self.moveDirection = 0


class RememberDir:
    """Context manager for remembering last opened directory.

    Parameters
    ----------
    window_instance: object
        reference to the window in which the context was initialized to center
        the directory selection dialog

    Attributes
    ----------
    _start_dir: str
        remembered directory if it could be read from file
    """

    _start_dir: str

    def __init__(self, window_instance: object) -> None:

        # keep reference to main window for dialog centering
        self.window_instance = window_instance

        # ensure directory for storing file exists
        makedirs(path.dirname(DIR_FILE), exist_ok=True)

    def get_dir(self) -> str:
        """Shows a folder selection dialog with the last visited dir as root.

        After the user has selected directory it is remenbered, returned to
        user and upon context exit saved to file.
        """
        self._start_dir = QFileDialog.getExistingDirectory(
            self.window_instance, "Open Folder", self._start_dir)

        return self._start_dir

    def __enter__(self):
        """ Load last visited directory from file.

        If the file could not be read, try to get music path on local PC
        """

        # load last opened dir
        try:
            f = open(DIR_FILE, "r")
        except FileNotFoundError:
            self._start_dir = get_music_path()
        else:
            self._start_dir = f.readline().strip()
            f.close()

        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """ On context exit try to save opened directory to file. """

        with open(DIR_FILE, "w") as f:
            f.write(self._start_dir)



"""Holds various html formats information extractors."""

import logging
import re  # lazy loaded
from typing import TYPE_CHECKING, List, Tuple

from wiki_music.constants import ORDER_NUMBER, TIME, TO_DELETE
from wiki_music.utilities import (NoTracklistException, SharedVars,
                                  normalize_caseless)

logging.getLogger(__name__)

__all__ = ["DataExtractors"]

if TYPE_CHECKING:
    from bs4 import BeautifulSoup


class DataExtractors:
    """Parse various table formats from wikipedia.

    Warnings
    --------
    This class is not ment to be instantiated, only inherited.
    """

    @classmethod
    def _from_table(cls, tables: List["BeautifulSoup"]
                    ) -> List[List[List[str]]]:
        """Extract a classic wkikipedia html table composed of 'td' and 'th'
        html tags.

        Parameters
        ----------
        tables: List[bs4.BeautifulSoup]
            each BeautifulSoup in list contains one htlm table

        Returns
        -------
        List[List[List[str]]]
            each emement in list is one parsed table, each table is a 2D array
            of strings representing rows and columns
        """

        data_collect = []
        for table in tables:

            # preinit list of lists
            rows = table.findAll("tr")
            row_lengths = [len(r.findAll(['th', 'td'])) for r in rows]
            ncols = max(row_lengths)
            nrows = len(rows)
            data = []

            for _ in rows:
                rowD = []
                for _ in range(ncols):
                    rowD.append('')
                data.append(rowD)

            # process html
            for i, row in enumerate(rows):
                cells = row.findAll(["td", "th"])

                for j, cell in enumerate(cells):
                    # lots of cells span cols and rows so lets deal with that
                    # row can probably span other subrows when there are
                    # subracks
                    cspan = int(cell.get('colspan', 1))
                    rspan = int(cell.get('rowspan', 1))

                    for k in range(rspan):
                        for l in range(cspan):
                            # sometimes cell is devided to subsells, so lets
                            # deal with this first
                            subcells = cell.find("div",
                                                 class_=re.compile("^hlist"))
                            if subcells:
                                c = ", ".join([s.text for s in
                                               subcells.findAll("li")])
                            else:
                                c = cell.text
                            data[i + k][j + l] += c

            for i in range(nrows):
                data[i] = list(filter(None, data[i]))
            data = list(filter(None, data))
            data_collect.append(data)

        return data_collect

    @classmethod
    def _html2python_list(cls, table: "BeautifulSoup") -> List[str]:
        """Converst html list to python list.

        Html list can be ordered <ol> or unordered <ul> its elements should be
        separated by <li> tags.

        Parameters
        ----------
        table: BeautifulSoup
            html object parsed by bs4

        Returns
        -------
        list
            each element represents on row in html list
        """

        try:
            rows = [ch.get_text() for ch in table.find_all("li")
                    if ch.string != "\n"]
        except AttributeError:
            return []
        else:
            return rows

    @classmethod
    def _from_list(cls, table: "BeautifulSoup") -> List[List[str]]:
        """Extract trackist formated as a html list with use of 'ol' and 'ul'
        tags.

        See also
        --------
        :meth:`_html2python_list`

        Parameters
        ----------
        table: BeautifulSoup
            html list containing the tracklist

        Returns
        -------
        List[List[str]]
            2D array representing table with rows and columns
        """

        rows = cls._html2python_list(table)

        if not rows:
            msg = ("No tracklist found!\n"
                   "It is probaly contained in some unknown format")
            SharedVars.warning(msg)
            raise NoTracklistException(msg)
        else:

            data: list = []
            for i, row in enumerate(rows):
                data.append([])

                try:
                    number = re.search(r"\d+\.", row).group()  # type: ignore
                except AttributeError:
                    number = str(i + 1)
                else:
                    rows[i] = row.replace(number, "")
                finally:
                    number = re.sub(r"^0|\.", "", number.strip())
                    data[-1].append(number)

                try:
                    time = re.search(TIME, row).group()  # type: ignore
                except AttributeError:
                    data[-1].append(row.strip())
                else:
                    rows[i] = re.sub(r"\s*–?-?\s*" + time, "", row)
                    rows[i] = row.replace(time, "")
                    rows[i] = re.sub(r"\"", "", row)
                    data[-1].append(row.strip())
                    data[-1].append(time)

            return [data]

    @classmethod
    def _get_track(cls, cell: str) -> Tuple[str, List[str]]:
        """Extract track and subtracks names from table cell.

        Parameters
        ----------
        cell: str
            table cell contining track and posiblly subtrack names

        Returns
        -------
        tuple
            first element is track name and second is a list of subtracks
        """

        # extract tracks and subtracks
        tracks = cell.split("\n")
        for j, t in enumerate(tracks):
            # delete " and strip whitespaces from ends
            t = t.replace("\"", "").strip()

            if j > 0:  # for subtracks
                # delete empty spaces, remove numbering if there is any
                t = re.sub(ORDER_NUMBER, "", t)

            # odstranenie zatvoriek s casom
            t = re.sub(TIME, "", t)
            # remove bonus in bracket behinf track name
            # TODO use TO_DELETE list from constats here
            tracks[j] = re.sub(TO_DELETE, "", t)

        return tracks[0], tracks[1:len(tracks)]

    @classmethod
    def _get_artist(cls, cell: str) -> List[str]:
        """Splits list of artists in tracklist table cell separated by , or &

        Parameters
        ----------
        cell: str
            string containing artists separated by delimites

        Returns
        -------
        List[str]
            list of artists
        """

        return [c.replace(",", "").strip() for c in re.split(",|&", cell)]

    @classmethod
    def _cut_out(cls, string: str, start: int, end: int) -> str:
        """Given a string cuts out part specified by pointers.

        Parameters
        ----------
        string: str
            string containing substring to be removed
        start: int
            pointer to the begining of the substring to be removed
        end: int
            pointer to the end of the substring to be removed

        Returns
        -------
        str
            copy of passed in string with part between pointers removed
        """
        return (string[:start] + string[end + 1:]).strip()



"""Module for handling flac tags."""

import logging
from collections import OrderedDict
from typing import Dict, Union

from mutagen.flac import FLAC, FLACNoHeaderError, Picture
from mutagen.id3 import PictureType

from .tag_base import TagBase

log = logging.getLogger(__name__)
log.debug("loading flac module")


class TagFlac(TagBase):
    """A low level implementation of tag handling for flac files. """
    __doc__ += TagBase.__doc__  # type: ignore

    _map_keys = OrderedDict([
        ("album", "ALBUM"),
        ("albumartist", "ALBUMARTIST"),
        ("artist", "ARTIST"),
        ("comment", "COMMENT"),
        ("composer", "COMPOSER"),
        ("date", "DATE"),
        ("discnumber", "DISCNUMBER"),
        ("genre", "GENRE"),
        ("lyrics", "LYRICS"),
        ("title", "TITLE"),
        ("tracknumber", "TRACKNUMBER"),
        ("picture", "COVERART")]
    )

    def _open(self, filename: str):
        """Function reading flac file to mutagen.flac.FLAC class. """

        try:
            self._song = FLAC(filename=filename)
        except FLACNoHeaderError:
            print("Cannot read FLAC tags")

    def _read(self) -> Dict[str, Union[str, bytearray]]:

        tags = dict()
        for key, value in self._map_keys.items():  # pylint: disable=no-member
            try:
                if key == "picture":
                    tag = self._song.pictures[0].data
                    continue
                else:
                    tag = self._song.tags[key]
                    if isinstance(tag, list):
                        tag = tag[0]
                    tag = str(tag).strip()
            except (KeyError, AttributeError):
                tag = ""
            finally:
                tags[value] = tag

        return tags

    def _write(self, tag: str, value: Union[str, bytearray]):

        if tag == "COVERART":

            pic = Picture()
            pic.type = PictureType.COVER_FRONT
            pic.mime = "image/jpeg"
            pic.data = value
            pic.desc = "Front Cover"

            # first we have to delete previous pictures and then write new
            self._song.clear_pictures()
            self._song.add_picture(pic)
        else:
            self._song.tags[self.reverse_map[tag]] = value



""" Holds constants that are used in parser. """
import re  # lazy loaded
from typing import Tuple, Pattern

__all__ = ["CONTENTS_IDS", "DEF_TYPES", "DELIMITERS", "COMPOSER_HEADER",
           "TO_DELETE", "UNWANTED", "NO_LYRIS", "ORDER_NUMBER", "WIKI_GENRES",
           "TIME", "PERSONNEL_SECTIONS"]

#: defines possible types of tracks that parser is able to extract
DEF_TYPES: Tuple[str, ...] = ("Instrumental", "Acoustic", "Acoustic Version",
                              "Orchestral", "Live", "Piano Version",
                              "Acoustic Folk Medley")
#: these tracks usually have no lyrics so no attempt to download them is made
NO_LYRIS = ("Instrumental", "Orchestral")
#: defines unwanted strings that will be deleted from track title, also helps
#: in extracting artists in brackets from track title
UNWANTED: Tuple[str, ...] = ("featuring", "feat.", "feat", "narration by",
                             "narration")
#: regex expressions that are deleted from track title
TO_DELETE: Pattern = re.compile(r"\( ?bonus ?(track)? ?\)", flags=re.I)
#: regex expresion that matches number with/out dot and any muber of spaces
ORDER_NUMBER: Pattern = re.compile(r"^ *\d+\.? *")
#: tuple with section headers from which personnel are extracted
PERSONNEL_SECTIONS: Tuple[str, ...] = ("personnel", "credits", "guests")
#: regex expression that matches wikipedia genres
WIKI_GENRES: Pattern = re.compile(r"/wiki/(?!Music_genre)", flags=re.I)
#: regex expression that matches time format e.g. (12:45)
TIME: Pattern = re.compile(r"\( *\d+\:\d+ *\)")
#: strings that are are used to extract composer name
DELIMITERS: Tuple[str, ...] = ("written by", "composed by", "lyrics by",
                               "music by", "arrangements by", "vocal lines by")
#: strings used to identify section headers in text and page contents
CONTENTS_IDS: Tuple[str, ...] = ("Track_listing", "Personnel", "Credits",
                                 "References")
#: table headers that are used to assign personnel to composers or to artists
#: list
COMPOSER_HEADER: Tuple[str, ...] = ("lyrics", "text", "music", "compose")



""" wiki_music submodule which provides all the GUI functionallity built with
support of all major python Qt bindings.
"""

import logging

# be sure to preserve import order!
from .custom_classes import (
    CustomQStandardItem, CustomQStandardItemModel, ImageTable, ImageWidget,
    NumberSortModel, RememberDir, ResizablePixmap, ResizableRubberBand,
    SelectablePixmap)
from .base import BaseGui
from .data_model import DataModel
from .cover_art import CoverArtSearch

__all__ = ["CustomQStandardItem", "CustomQStandardItemModel", "ImageTable",
           "ImageWidget", "NumberSortModel", "RememberDir", "ResizablePixmap",
           "ResizableRubberBand", "SelectablePixmap", "BaseGui",
           "CoverArtSearch", "DataModel"]

logging.getLogger(__name__)



"""Utility functions and classes used by parser."""

import collections  # lazy loaded
import logging
import os
import sys
import time  # lazy loaded
from threading import Lock, Thread
from typing import (TYPE_CHECKING, Any, Callable, Dict, Generator, List,
                    Optional, Tuple, Union)

import fuzzywuzzy.fuzz as fuzz  # lazy loaded
import yaml  # lazy loaded

from wiki_music.constants.colors import GREEN, RESET

from .utils import normalize

logging.getLogger(__name__)

if TYPE_CHECKING:
    from logging import Logger

__all__ = ["ThreadWithTrace", "NLTK", "bracket", "write_roman",
           "normalize", "normalize_caseless", "caseless_equal",
           "caseless_contains", "count_spaces", "yaml_dump",
           "complete_N_dim", "replace_N_dim", "delete_N_dim",
           "ThreadPool", "yaml_load", "ThreadWithReturn"]


class ThreadWithTrace(Thread):
    """Subclass of threading.thread, sets trace to thread by means of which
    it can be later killed from outside

    Parameters
    ----------
    args: Any
        same as threading.Thread
    kwargs: Any
        same as threading.Thread

    References
    ----------
    https://www.geeksforgeeks.org/python-different-ways-to-kill-a-thread/
    """

    def __init__(self, *args, **keywords) -> None:
        Thread.__init__(self, *args, **keywords)
        self.killed: bool = False

    def start(self):
        self.__run_backup = self.run
        self.run = self.__run
        Thread.start(self)

    def __run(self):
        sys.settrace(self.globaltrace)
        self.__run_backup()
        self.run = self.__run_backup

    def globaltrace(self, frame, event, arg):
        if event == 'call':
            return self.localtrace
        else:
            return None

    def localtrace(self, frame, event, arg):
        if self.killed:
            if event == 'line':
                raise SystemExit()
        return self.localtrace

    def kill(self):
        self.killed = True


class _NltkMeta(type):
    """Metaclass which defines thread safe nltk property. Only for use in
    NLTK class.

    See Also
    --------
    :class:`NLTK`
    """

    @property
    def nltk(cls):
        """Thread safe property which holds reference to nltk lib.

        :type: nltk
        """
        with cls._lock:
            return cls._nltk


class NLTK(metaclass=_NltkMeta):
    """A thread safe nltk importer. Will make other threads wait if they want
        to access nltk until it is imported.
    """

    _import_running: bool = False
    # nltk class attribute is provided by metaclass
    _nltk = None
    _lock: Lock = Lock()
    
    @classmethod
    def run_import(cls, logger: "Logger"):
        """Import nltk in separate thread and assign it to class attribute.
        
        Parameters
        ----------
        logger: logging.Logger
            instance of a logger to log import messages
        """

        def imp():
            with cls._lock:
                logger.debug("import nltk")
                try:
                    import nltk
                    cls._nltk = nltk
                except ImportError as e:
                    logger.debug(f"failed to import nltk: {e}")
                else:
                    logger.debug("import nltk done")

        if not cls._import_running:

            cls._import_running = True
            # load NLTK in separate thread
            Thread(target=imp, name="ImportNLTK").start()
        else:
            log.debug("nltk import already running")


class ThreadWithReturn(Thread):
    """ Subclass of python threading.Thread which can return result of
    running function. The result is return by calling the Thread.join()
    method.

    Parameters
    ----------
    args: Any
        same as threading.Thread
    kwargs: Any
        same as threading.Thread

    References
    ----------
    https://stackoverflow.com/questions/6893968/how-to-get-the-return-value-from-a-thread-in-python

    See Also
    --------
    :class:`ThreadPool`
    """

    def __init__(self, *args, **kwargs) -> None:
        super(ThreadWithReturn, self).__init__(*args, **kwargs)

        self._return: Any = None

    def run(self):
        """ Override standard threading.Thread.run() method to store
        running function return value.
        """

        if self._target is not None:
            self._return = self._target(*self._args, **self._kwargs)

    def join(self, timeout:Optional[float] = None) -> Any:
        """ Override standard threading.Thread.join() method to return
        running function return value.
        """

        super(ThreadWithReturn, self).join(timeout=timeout)

        return self._return


class ThreadPool:
    """ Spawns pool of threads to excecute function. If the list of arguments
    contains only one tuple, run the function in the calling thread to
    avoid unnecessary overhead as a result of spawning a new thread.

    Parameters
    ----------
    target: Callable
        callable that each thread should run
    args: List[tuple]
        each tuple in list contains args for one thread running target
    
    See Also
    --------
    :class:`ThreadWithReturn`
    """

    def __init__(self, target:Callable[..., list] = lambda *args: [],
                 args:List[tuple] = [tuple()]) -> None:

        self._args = args
        self._target = target

    def run(self, timeout: Optional[float] = 60) -> list:
        """ Starts the execution of threads in pool. returns after all threads
        join() metod has returned.

        Parameters
        ----------
        timeout: Optional[float]
            timeout after which waiting for results will be abandoned

        Returns
        -------
        list
            list of returned values from the functions run by the ThreadPool
        """

        if len(self._args) == 1:
            return self._target(*self._args)
        else:
            threads: List[ThreadWithReturn] = []

            for i, a in enumerate(self._args):
                threads.append(ThreadWithReturn(target=self._target, args=a,
                                                name=f"ThreadPoolWorker-{i}"))
                threads[-1].daemon = True
                threads[-1].start()
            
            for i, l in enumerate(threads):
                threads[i] = l.join(timeout=timeout)

            return threads
    

def bracket(data: List[str]) -> List[str]:
    """ Puts elements of the list in brackets.\n

    Parameters
    ----------
    data: List[str]
        input data

    Returns
    -------
    List[str]
        list of strings with brackets at the ends
    """

    data_tmp = []
    for d in data:
        if d:
            data_tmp.append("({})".format(d))
        else:
            data_tmp.append("")

    return data_tmp


def write_roman(num: Union[int, str]):
    """ Convert integer to roman number 
    Parameters
    ----------
    num: int
        integer to convert to roman number

    References
    ----------
    https://stackoverflow.com/questions/52554702/roman-numerals-to-integers-converter-with-python-using-a-dictionary

    Returns
    -------
    str
        roman number converted from integer
    """

    roman_numerals = [
        ('M', 1000),
        ('CM', 900),
        ('D', 500),
        ('CD', 400),
        ('C', 100),
        ('XC', 90),
        ('L', 50),
        ('XL', 40),
        ('X', 10),
        ('IX', 9),
        ('V', 5),
        ('IV', 4),
        ('I', 1)
    ]

    num = str(num)

    ix = 0
    result = 0
    while ix < len(num):
        for k, v in roman_numerals:
            if num.startswith(k, ix):
                result += v
                ix += len(k)
                break
        else:
            raise ValueError('Invalid Roman number.')
    return result


def normalize_caseless(text: str) -> str:
    """ NFKD casefold string normalization 

    Parameters
    ----------
    text: str
        string to normalize

    Returns
    -------
    str
        normalized caseless version of string
    """
    return normalize(text).casefold()


def caseless_equal(left: str, right: str) -> bool:
    """ Check for normalized string equality 

    Parameters
    ----------
    left: str
        string to compare
    right: str
        string to compare

    Returns
    -------
    bool
        true if normalized caseless strings are equal

    See also
    --------
    :func:`normalize_caseless`
    """
    return normalize_caseless(left) == normalize_caseless(right)


def caseless_contains(string: str, in_text: str) -> bool:
    """ Check if string is contained in text.\n

    Parameters
    ----------
    string: str
        string to search for
    in_text: str
        string to search in

    Returns
    -------
    bool
        True if string is in text

    See also
    --------
    :func:`normalize_caseless`
    """

    if normalize_caseless(string) in normalize_caseless(in_text):
        return True
    else:
        return False


def count_spaces(*lists: Tuple[List[str], ...]) -> Tuple[List[str], int]:
    """ Counts max length of elements in list and croesponding spaces for
    each item to fit that length.

    Parameters
    ----------
    lists: Tuple[List[str]]
        data on which to count spaces, all lists in tuple must be of
        same length

    Returns
    -------
    tuple
        list of number os apces to append to list elements to make them span
        max length
    """

    transposed: List[List[str]] = list(map(list, zip(*lists)))
    max_length: int = 0
    spaces: List[str] = []

    for trans in transposed:
        length = sum([len(t) for t in trans])
        if length > max_length:
            max_length = length

    for trans in transposed:
        spaces.append(" " * (max_length - sum([len(t) for t in trans]) + 8))

    return spaces, max_length


def yaml_dump(dict_data: List[Dict[str, str]], save_dir: str):
    """ Save yaml file to disk. Each dictionary in list contains tags 
    of one album track

    Parameters
    ----------
    dict_data: List[Dict[str, str]]
        list of dictionarie to save to disk
    save_dir: str
        directory to save to
    """

    _path = os.path.join(save_dir, "database.yaml")
    print(GREEN + "\nSaving YAML file: " + RESET + _path + "\n")
    with open(_path, "w") as outfile:
        yaml.dump(dict_data, outfile, default_flow_style=False)


def yaml_load(yml_file: str) -> List[dict]:
    """ Loads yaml format file to dictionary. 
    
    Parameters
    ----------
    yml_file: str
        path to yml file

    Returns
    -------
    List[dict]
        list of loaded dictionaries
    """

    with open(yml_file, "r") as infile:
        return yaml.full_load(infile)


def _find_N_dim(array: Union[list, str], template: str
               ) -> Optional[Union[list, str]]:
    """ Recursive helper function with two nested list as input. array is
    traversed and its elements are fuzzy tested if they match expresion in 
    template

    Parameters
    ----------
    array: list
        argument is a nested list of strings in which we want to find
        element.
    template: str
       string to find in the nested list

    See also
    --------
    :func:`complete_N_dim`
    :func:`replace_N_dim`
    """

    if isinstance(array, list):
        for a in array:
            ret = _find_N_dim(a, template)
            if ret is not None:
                return ret
    else:
        # length check is to stop shortz words replacing long
        if len(array) > len(template):
            if fuzz.token_set_ratio(template, array) > 80:
                return array


def complete_N_dim(to_complete: list, to_find: list):
    """ Recursive function with two list as input, one list contains incomplete
    versions of strings and the other has full versions. Lists can be nested.
    both are then traversed and the strings in the first list are completed 
    with strings from the second list. Changes are made in place.

    Parameters
    ----------
    to_complete: list
        argument is a list which elements are not complete. Changes to this
        list are made in-place.
    to_find: list
       argument is the list which contains full strings. All uncomplete
       strings in to_complete will be replaced by their longer version
       from to_find

    See also
    --------
    :func:`_find_N_dim`
    """

    if isinstance(to_complete, list):
        for i, _ in enumerate(to_complete):
            ret = complete_N_dim(to_complete[i], to_find)
            if ret is not None:
                to_complete[i] = ret
    else:
        return _find_N_dim(to_find, to_complete)


def replace_N_dim(to_replace: list, to_find: str):
    """ Recursive function with nested list as input. The nested list elements
    are traversed and defined expresion is replaced by empty string in each
    element. Changes are made in place.

    Parameters
    ----------
    to_replace: list
        argument is a list which elements are not contain unwanted string.
    to_find: str
       the unwanted string to find in nested list

    See also
    --------
    :func:`_find_N_dim`
    """

    if isinstance(to_replace, list):
        for i, _ in enumerate(to_replace):
            ret = replace_N_dim(to_replace[i], to_find)
            if ret is not None:
                to_replace[i] = ret
    else:
        return to_replace.replace(to_find, "").strip()


def delete_N_dim(to_delete: list, to_find: list) -> list:  # type: ignore
    """ Recursive function with nested list as input. The nested list elements
    are traversed and each that is equal to one of the elements in to_find list
    is deleted. Changes are made in place.

    Parameters
    ----------
    to_replace: list
        argument is a nested list which contains unwanted elements.
    to_find: str
       list of unwanted elements
    """

    if to_delete:
        if isinstance(to_delete[0], list):
            for i, td in enumerate(to_delete):
                to_delete[i] = delete_N_dim(td, to_find)
        else:
            return [td for td in to_delete if td not in to_find]
    else:
        return []



"""Extractor for azlyrics.com."""

import logging
import re

from ..extractor import LyricsExtractor
from ..models.lyrics import Lyrics
from ..models import exceptions

log = logging.getLogger(__name__)


class AZLyrics(LyricsExtractor):
    """Class for extracting lyrics."""

    name = "AZLyrics"
    url = "https://www.azlyrics.com/"
    display_url = "azlyrics.com"

    @staticmethod
    def extract_lyrics(url_data, song, artist):
        """Extract lyrics."""
        url_data.headers = {"user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:59.0) Gecko/20100101 Firefox/59.0"}
        bs = url_data.bs

        center = bs.body.find("div", {"class": "col-xs-12 col-lg-8 text-center"})
        
        if not center:
            # 503 Service Temporarily Unavailable
            if re.search(r"50[0-9]", bs.find("title").text):
                raise exceptions.NotAllowedError
            # if song is not present AZLyrics returns title page
            elif bs.find("title").text == "AZLyrics - Song Lyrics from A to Z":
                raise exceptions.NoLyrics
            # if we get to the album page instead of song page
            else:
                # find tracklist
                tracklist = bs.find("div", {"id", "listAlbum"})
                # find song url
                url_data.url = tracklist.find("a", text=song.title())["href"]
                # call method again
                return AZLyrics.extract_lyrics(url_data, song, artist)
        
        lyrics = center.find("div", {"class": None}).text

        lyrics = re.sub(r"<br>", " ", lyrics)
        lyrics = re.sub(r"<i?>\W*", "[", lyrics)
        lyrics = re.sub(r"\W*</i>", "]", lyrics)
        lyrics = re.sub(r"\&quot\;", "\"", lyrics)
        lyrics = re.sub(r"</div>", "", lyrics)

        title = center.find("h1").text.strip()[1:-8]
        artist = bs.select_one("div.lyricsh h2 b").string[:-7]
        lyrics = lyrics.strip()

        return Lyrics(title, lyrics, artist=artist)



"""Extractor for genius.com."""

import logging
from datetime import datetime

from ..extractor import LyricsExtractor
from ..models.lyrics import Lyrics
from ..models import exceptions

log = logging.getLogger(__name__)


class Genius(LyricsExtractor):
    """Class for extracting lyrics."""

    name = "Genius"
    url = "https://genius.com/"
    display_url = "genius.com"

    @staticmethod
    def extract_lyrics(url_data, song, artist):
        """Extract lyrics."""
        bs = url_data.bs

        try:
            lyrics_window = bs.find_all("div", {"class": "lyrics"})[0]
        except IndexError:
            # Oops! Page not found message
            if bs.find("h1", {"class": "render_404-headline"}):
                raise exceptions.NoLyrics
            # if album name == song name and we end up in album page
            else:
                # find tracklist
                tracklist = bs.find(
                    "div", {"class": "track_listing track_listing--columns"})
                # extract new url
                url_data.url = tracklist.find("a", {"title": song})["href"]
                # call extraction method again
                return Genius.extract_lyrics(url_data, song, artist)

        lyrics = lyrics_window.text.strip()

        title = bs.find("h1", attrs={"class": "header_with_cover_art-primary_info-title"}).text
        artist = bs.select_one("a.header_with_cover_art-primary_info-primary_artist").string
        try:
            date_str = bs.find(text="Release Date").parent.find_next_sibling("span").string
        except AttributeError:
            date_str = "January 1, 2000"
        release_date = datetime.strptime(date_str, "%B %d, %Y")

        return Lyrics(title, lyrics, artist=artist, release_date=release_date)



import logging
import re  # lazy loaded
from os import path
from typing import Dict, Optional, Union

import mutagen  # lazy loaded
from wiki_music.constants.colors import GREEN, RED, RESET, YELLOW
from wiki_music.constants.tags import TAGS
from wiki_music.library import tags_handler
from wiki_music.utilities import SharedVars, exception

__all__ = ["read_tags", "write_tags", "supported_tags"]

log = logging.getLogger(__name__)


def supported_tags():
    """Print a list of wiki_music supported tags."""

    print("wiki_music supports these tags:")
    print(", ".join(TAGS))


@exception(log)
def write_tags(data: Dict[str, Union[str, float, list, bytearray]]):
    """Convenience function which takes care of writing data to tags.

    See also
    --------
    :const:`wiki_music.constants.tags.TAGS`
        for list of supported tags
    :mod:`wiki_music.library.tags_handler`
        low level implemetation of tags handling built on mutagen library

    Parameters
    ----------
    data: dict
        containes dictionary of tag names and coresponding values
    """

    def write(tag, value=None):

        if value is None:
            value = data[tag]

        if isinstance(value, list):
            if len(value) == 1:
                value = value[0]
            else:
                value = ", ".join(value)
        if isinstance(value, (int, float)):
            value = str(value)

        song.tags[tag] = value

    if not data["FILE"]:
        print(f"{data['TITLE']} {data['type']}" + YELLOW +
              "does not have matching file!")
        return

    try:
        song = tags_handler.File(data["FILE"])
    except mutagen.MutagenError as e:
        print(f'Couldn´t open file {data["FILE"]} for writing')
        SharedVars.exception(e)
        log.exception(e)
    else:

        print(GREEN + "Writing tags to:" + RESET, data["FILE"])

        # preprocess data
        if isinstance(data["ARTIST"], list):
            data["ARTIST"].append(data["ALBUMARTIST"])
        elif isinstance(data["ARTIST"], str):
            data["ARTIST"] = [data["ALBUMARTIST"], data["ARTIST"]]
        else:
            raise NotImplementedError("Unsupported data type for "
                                        "ARTIST tag")

        data["ARTIST"] = sorted(list(set(data["ARTIST"])))
        if data["TYPE"]:
            data["TITLE"] = f"{data['TITLE']} {data['TYPE']}"

        # write tags, except file and type,
        # these don't belong in music tags
        for t in TAGS:
            if t not in ("FILE", "TYPE"):
                write(t)

        write("COMMENT", "")

        try:
            song.save()
        except (mutagen.MutagenError, TypeError, ValueError) as e:
            print(f'Couldn´t save file {data["FILE"]}')
            SharedVars.exception(e)
            log.exception(e)
        else:
            print(GREEN + "Tags written succesfully!")


@exception(log)
def read_tags(song_file: str) -> Dict[str, Union[str, list, bytearray]]:
    """ Convenience function which takes care of reading tags from file.
    Abstracts away from low level mutagen API. If no tags are read, function
    can guess track title from file name, assumming some decent formating.

    See also
    --------
    :const:`wiki_music.constants.tags.TAGS`
        for list of supported tags
    :mod:`wiki_music.library.tags_handler`
        low level implemetation of tags handling built on mutagen library

    Parameters
    ----------
    song_file: str
        path to song on disk

    Returns
    -------
    dict
        dictionary of tag labels with coresponding values if the file could be
        loaded. If not then the dictionary is empty
    """

    try:
        song = tags_handler.File(song_file)
    except mutagen.MutagenError as e:
        print(RED + "Error in reading file: " + RESET + song_file)
        SharedVars.exception(e)
        log.exception(e)
        return {}
    else:
        # convert selective dict to normal dict, selectivness is needed only
        # for writing tags
        tags = dict(song.tags)

        if not tags["TITLE"]:
            f = path.splitext(path.basename(song_file))[0]
            # match digits, zero or one whitespace characters,
            # zero or one dot, zero or one dash,
            # zero or one whitespace characters
            tags["TITLE"] = re.sub(r"\d\s?\.?-?\s?", "", f).strip()

        for s in ("ARTIST", "COMPOSER"):
            tags[s] = sorted(re.split(r",\s?", tags[s]))

        return tags



""" This module houses classes that are responsible for search and retrieval
of cover art picture.
"""

import logging
import queue
from os import path
from threading import Thread
from typing import Dict, Iterable, Optional, Tuple, Union

from wiki_music.constants import (ASPECT_RATIOS, COVER_ART_EDIT_UI,
                                  COVER_ART_SEARCH_UI)
from wiki_music.external_libraries.google_images_download import (  # lazy loaded
    google_images_download, google_images_download_offline)
from wiki_music.gui_lib import BaseGui, ImageTable, SelectablePixmap
from wiki_music.gui_lib.qt_importer import (QDialog, QHBoxLayout, QIcon,
                                            QMessageBox, QSize, Qt, QTimer,
                                            QWidget, Signal, uic)
from wiki_music.utilities import (MultiLog, SharedVars, comp_res, exception,
                                  get_icon, get_image, get_image_size)
                                  # , send_to_clipboard)

log = logging.getLogger(__name__)
log.debug("coverart imports done")


class PictureContainer(SelectablePixmap):
    """Holds the selected picture that is being edited. The raw picture is
    stored in bytes data. Which are the loaded by `PIL` for manipulation
    or by GUI to display.

    Warnings
    --------
    This class has the ability to select parts of picture with a rubberband
    so it should be used only for editing pictures.

    Parameters
    ----------
    bytes_image: bytearray
        the picture that will be edited in bytes format
    log: :class:`wiki_music.utilities.util.MultiLog`
        logger instance of creating class

    Attributes
    ----------
    bytes_image_orig: bytearray
        original image in byres format
    bytes_image_edit: bytearray
        original image with applied compession
    bytes_image_crop: Optional[bytearray]
        cropped version of original image without compression
    bytes_image_resz: Optional[bytearray]
        resized version of original image without compression
    log: :class:`wiki_music.utilities.util.MultiLog`
        class logger
    disksizeChanged: Signal(str)
        signal that is emited whenever the disk size of image changes by
        compresion, resize or crop
    dimensionsChanged: Signal(int, int)
        signal emited on picture resize or crop

    See also
    --------
    :class:`wiki_music.gui_lib.custom_classes.SelectablePixmap`
        inhereted class which implements the ability to select with rubberband
    """

    bytes_image_orig: bytearray
    bytes_image_edit: bytearray
    bytes_image_crop: Optional[bytearray]
    bytes_image_resz: Optional[bytearray]
    log: MultiLog

    disksizeChanged: Signal = Signal(str)
    dimensionsChanged: Signal = Signal(int, int)

    def __init__(self, bytes_image: bytearray, log: MultiLog) -> None:

        super().__init__(bytes_image)

        self.bytes_image_orig = bytes_image
        self.bytes_image_edit = bytes_image
        self.bytes_image_crop = None
        self.bytes_image_resz = None

        self.resize(800, 800)
        self.log = log

    @property
    def current_image_c(self) -> bytearray:
        """ Get image to compress.

        Returns
        -------
        bytearray
            the appropriate image version
        """

        # if we have cropped image, use that
        if self.bytes_image_crop:
            return self.bytes_image_crop
        # if no cropped image is present use resized
        elif self.bytes_image_resz:
            return self.bytes_image_resz
        # is we have none of those use the original
        else:
            return self.bytes_image_orig

    @property
    def current_image_r(self) -> bytearray:
        """ get image to resize.
        
        Returns
        -------
        bytearray
            the appropriate image version
        """

        # if some resized version is present use that, otherwise use original
        if self.bytes_image_crop:
            return self.bytes_image_crop
        else:
            return self.bytes_image_orig

    def compress_image(self, quality: int):
        """ Apply defined compresion to image and show result.

        Parameters
        ----------
        quaility: int
            value between 1 and 99 setting the compresion level

        See also
        --------
        :func:`wiki_music.utilitie.gui_utils.comp_res`
            compresses image using PIL
        """

        self.bytes_image_edit = comp_res(self.current_image_c, quality)
        self.update_pixmap(self.bytes_image_edit)

        size = get_image_size(self.bytes_image_edit)
        self.disksizeChanged.emit(size)

        self.log.info(f"Compressed cover art  to: {size}Kb")

    def resize_image(self, x: int, y: int, quality: int):
        """ Resize image to set dimensions, then apply defined compresion and
        show result.

        Parameters
        ----------
        quaility: int
            value between 1 and 99 setting the compresion level
        x: int
            horizontal image dimension in pixels
        y: int
            vertical image dimension in pixels

        See also
        --------
        :func:`wiki_music.utilitie.gui_utils.comp_res`
            compresses and resizes image using PIL
        """

        self.bytes_image_edit = comp_res(self.current_image_r, quality, x, y)
        self.bytes_image_resz = self.bytes_image_edit
        self.update_pixmap(self.bytes_image_resz)

        self.log.info(f"Cover art resized to: {x}x{y}")
        self.disksizeChanged.emit(get_image_size(self.bytes_image_edit))

    def crop_image(self, quality: int):
        """ Crops image to part selected by rubberband, then applies set level
        of compresin and shows image in GUI.

        Parameters
        ----------
        quaility: int
            value between 1 and 99 setting the compresion level

        See also
        --------
        :func:`wiki_music.utilitie.gui_utils.comp_res`
            compresses image using PIL
        """

        # TODO wrong image part displayed after crop
        crop_pixmap = self.pixmap().copy(self.currentQRubberBand.geometry())
        self.bytes_image_crop = self._pixmap2bytes(crop_pixmap)
        self.compress_image(quality)
        self.cancel_selection()
        self.update_pixmap(self.bytes_image_crop)

        self.dimensionsChanged.emit(*self.image_dims)

        self.log.info("Cover art cropped to: {}x{}".format(*self.image_dims))

    def save_image(self, path: str):
        """ Saves the most current version of the image to file with name
        Folder.jpg
        """

        with open(path, 'wb') as f:
            f.write(self.bytes_image_edit)

    # TODO currently not used
    # def send2clipboard(self):
    #    """ Copies the image to clipboard """
    #    send_to_clipboard(self.bytes_image_edit)


class SearchDialog(QDialog):
    """ Manages the search dialog window, displays found picture thumbnails and
    houses some basic search controls.

    Parameters
    ----------
    query: str
        search phrase passed to google images download

    Attributes
    ----------
    table: :class:`wiki_music.gui_lib.custom_classes.ImageTable`
        special subclass of Qt ImageTable that can hold pictures in cells and
        can add new pictures on the fly
    pictureClicked: Signal(int)
        signal emited when table cell is clicked

    See also
    --------
    :class:`wiki_music.ui.cover_art_base.Ui_cover_art_search`
        this is the template class for this class layout
    :class:`wiki_music.gui_lib.custom_classes.ImageTable`
        this is the class ttaht manages the table itself
    """

    table: ImageTable

    pictureClicked: Signal = Signal(int)

    def __init__(self, query: str) -> None:

        # QDialog init
        super().__init__()
        # setup ui from template
        uic.loadUi(COVER_ART_SEARCH_UI, self)

        self.setWindowTitle("Wiki Music - Cover Art search")
        self.setWindowIcon(QIcon(get_icon()))

        # set initial values
        self.progressBar.setRange(0, 100)
        self.progressBar.setValue(0)
        self.query.setText(query)

        # create table for image thumbnails
        self.table = ImageTable()
        self.verticalLayout.insertWidget(0, self.table)

        # forward table signal
        self.table.cellClicked.connect(self.table_cell_clicked)

        # connect cancel button
        self.cancel_button.clicked.connect(self.done)

    def table_cell_clicked(self, row: int, col: int):
        """ Emits signal when cell in table is clicked. From the row and
        column position calculates the picture position in the list.

        Parameters
        ----------
        row: int
            row pocition of clicked cell
        col: int
            column position of clicked cell
        """
        self.pictureClicked.emit((self.max_columns + 1) * row + col)

    @property
    def max_columns(self) -> int:
        """ Get fixed predefined number of row columns.

        Returns
        -------
        int:
            predefined number of table columns
        """
        return self.table.MAX_COLUMNS

    def add_pic(self, dimension: str, thumbnail: bytearray):
        """ Add new picture to the table.
        
        Parameters
        ----------
        dimension: str
            formated string with image dimension and size in Kb
        thumbnail: bytearray
            raw image thumbnail in bytes format 
        """
        self.table.add_pic(dimension, thumbnail)


class PictureEdit(QDialog):
    """ Manages picture editor window, with cropping, resizing and compressing
    abilities.

    Parameters
    ----------
    dimensions: Tuple[int, int]
        dimensions of the image being edited
    clicked_image: bytearray
        the image to edit in bytes format
    log: :class:`wiki_music.utilities.utils.MultiLog`
        main GUI class logger
    save_dir: str
        directory in which the edited image will be saved if the option is
        chosen

    Attributes
    ----------
    picture: PictureContainer
        class holding the picture and its metods
    original_ratio: float
        save original ratio of the picture
    editingFinished: Signal
        signal to emit the edited picture when choice is accepted
    """

    log: MultiLog
    picture: PictureContainer
    original_ratio: float

    editingFinished: Signal = Signal(bytearray)

    def __init__(self, dimensions: Tuple[int, int], clicked_image: bytearray,
                 log: MultiLog, save_dir: str) -> None:

        # QDialog init
        super().__init__()
        # setup ui from template
        uic.loadUi(COVER_ART_EDIT_UI, self)

        self.log = log
        self.save_dir = save_dir

        self.setWindowTitle("Edit cover art")
        self.setWindowIcon(QIcon(get_icon()))
        self.picture = PictureContainer(clicked_image, self.log)

        # variables
        self.original_ratio = dimensions[0] / dimensions[1]

        self.ca_dropdown.insertItems(1, ASPECT_RATIOS)
        self.size_box.setReadOnly(True)
        self.cancel_crop.setEnabled(False)

        self.size_spinbox_X.setValue(dimensions[0])
        self.size_spinbox_Y.setValue(dimensions[1])
        self.compresion_spinbox.setValue(95)
        self.compresion_slider.setValue(95)
        self.size_box.setText(get_image_size(clicked_image))
        self.picture_frame.addWidget(self.picture)

        self.__setup_overlay__()

    def __setup_overlay__(self):
        """ Connect all the dialog elements to coresponding signals. """

        # editor signals
        self.picture.selectionActive.connect(self.cancel_crop.setEnabled)
        self.picture.disksizeChanged.connect(self.size_box.setText)
        self.picture.dimensionsChanged.connect(self.set_image_dims)

        # crop signals
        self.apply_crop.clicked.connect(
            lambda x: self.picture.crop_image(self.compresion))
        self.cancel_crop.clicked.connect(
            lambda x: self.picture.cancel_selection())
        self.ca_dropdown.editTextChanged.connect(self._get_crop_ratio)

        # resize signals
        self.size_spinbox_X.valueChanged.connect(
            lambda size: self._aspect_ratio_sync(size, "x"))
        self.size_spinbox_Y.valueChanged.connect(
            lambda size: self._aspect_ratio_sync(size, "y"))
        self.apply_ratio.clicked.connect(
            lambda x: self.picture.resize_image(*self.image_dims,
                                                self.compresion))

        # compression signals
        self.compresion_spinbox.valueChanged.connect(
            lambda comp_value: self._get_compressed(comp_value, "spinbox"))
        self.compresion_slider.valueChanged.connect(
            lambda comp_value: self._get_compressed(comp_value, "slider"))

        # on finished editing emit signal with edited image
        self.accepted.connect(
            lambda: self.editingFinished.emit(self.picture.bytes_image_edit))

    @property
    def image_dims(self) -> Tuple[int, int]:
        """ Edited image dimensions.

        :type: Tuple[int, int]
        """
        return self.size_spinbox_X.value(), self.size_spinbox_Y.value()

    def set_image_dims(self, x: int, y: int):
        """ passes current image dimension to spinboxes.

        Parameters
        ----------
        x: int
            horizontal dimension in pixels
        y: int
            vertical dimension in pixels
        """

        self.size_spinbox_X.setValue(x)
        self.size_spinbox_Y.setValue(y)

    @property
    def clipboard(self) -> bool:
        """ Clipboard checkbox state.

        :type: bool
        """
        return self.ca_clipboard.isChecked()

    @property
    def save_file(self) -> bool:
        """ Save to disk checkbox state.

        :type: bool
        """
        return self.ca_file.isChecked()

    @property
    def preserve_ratio(self) -> bool:
        """ Preserve ration checkbox state.

        :type: bool
        """
        return self.ca_ratio.isChecked()

    @property
    def compresion(self) -> int:
        """ Current compresion slider and spinbox value.

        :type: int
        """
        return self.compresion_spinbox.value()

    @exception(log)
    def _get_crop_ratio(self, value: str):
        """ Sets the desired aspect ratio for cropping

        See also
        --------
        :meth:`PictureContainer.set_aspect_ratio`
            method activated to apply the aspect ratio to picture

        Paremeters
        ----------
        value: str
            ratio value can be one of:
            :const:`wiki_music.constants.gui_const.ASPECT_RATIOS`

        Raises
        ------
        ValueError
            if the passed `value` is not a valid ratio format
        """

        crop_ratio: Optional[Iterable[float]]

        if value.lower() == "preserve":
            crop_ratio = (self.original_ratio, 1)
        elif value.lower() == "free":
            crop_ratio = None
        else:
            try:
                crop_ratio = [int(i) for i in value.split(":")]
            except ValueError:
                raise ValueError(f"{value} is not a valid ratio format")

        self.picture.set_aspect_ratio(crop_ratio)

    def _aspect_ratio_sync(self, dim: int, activated: str):
        """ Synchronize the aspect ratio spinboxes. Employs complex logic to
        avoid recursion.

        Parameters
        ----------
        dim: int
            the current value in spinbox which activated this method
        activated: str
            x or y the spinbox which activated this methos

        """

        x, y = self.image_dims

        if self.preserve_ratio:
            # TODO find some nicer way to avoid recursion
            if activated == "x":
                if int(self.original_ratio * y) == x:
                    return
                y = int((1 / self.original_ratio) * dim)
                self.size_spinbox_Y.setValue(y)
            elif activated == "y":
                if int((1 / self.original_ratio) * x) == y:
                    return
                x = int(self.original_ratio * dim)
                self.size_spinbox_X.setValue(x)
        else:
            self.original_ratio = x / y

    def _get_compressed(self, comp_value: int, activated: str):
        """ Compresses the image to desired value.

        Parameters
        ----------
        comp_value: int
            compression value from interval (1, 99)
        activated: str
            where to get the value from, spinbox or slider

        See also
        --------
        :meth:`PictureContainer.compress_image`
            method activated to apply the compression to picture
        """

        if activated == "spinbox":
            self.compresion_slider.setValue(comp_value)
        elif activated == "slider":
            self.compresion_spinbox.setValue(comp_value)

        self.picture.compress_image(comp_value)

    def picture_save(self):
        """ Saves picture to Folder.jpg in directory with
        currently edited files.

        Warnings
        --------
        Copying to clipboard is disabled for now

        See also
        --------
        :meth:`PictureContainer.save_image`
            method activated to save picture to file
        :meth:`PictureContainer.send2clipboard`
            method activated to paste picture to clipboard
        """

        if self.save_file:
            self.log.info("Cover art saved to file")
            self.picture.save_image(path.join(self.save_dir, "Folder.jpg"))
        if self.clipboard:
            self.log.info("Cover art copied to clipboard")
            # TODO this is not used for now see discussion in gui_utils module
            # self.picture.send2clipboard()


class CoverArtSearch(BaseGui):
    """ Main class that handles cover art search, calls all the apropriate
    methods. First initialize google_images_download in a separate thread,
    then initializes Search dialog window. After that it reads downloaded image
    thumbnails and displays them inn the dialog.

    Attributes
    ----------
    image_search_thread: Thread
        thread that runs the goolgle_images download in the background
    search_dialog: SearchDialog
        reference to dialog window showing the downloaded pictures in table
    picture_editor: PictureEdit
        reference to dialog window that handles selected image editing

    See also
    --------
    :mod:`wiki_music.external_libraries.google_images_download`
        submodule that handles downloading cover art images
    :class:`wiki_music.gui_lib.cover_art.SearchDialog`
        class that shows dialog downloaded cover art images and some basic
        search controls
    :class:`wiki_music.gui_lib.cover_art.PictureEdit`
        handles editing of downloaded pictures
    """

    _max_count: int
    images: Dict[str, Union[bytes, str, tuple]]
    image_search_thread: Thread
    search_dialog: SearchDialog
    picture_editor: PictureEdit

    @exception(log)
    def cover_art_search(self):
        """ Method that takes care of cover art search, download and edit.

        See also
        --------
        :meth:`_async_loader`
            method that loads results to GUI
        :meth:`_select_picture`
            method that stats picture edit dialog
        """
        self.images = []

        if SharedVars.offline_debbug:
            self.gimd = google_images_download_offline.googleimagesdownload()
            self._max_count = 5
        else:
            if not all((self.ALBUMARTIST, self.ALBUM)):
                QMessageBox(QMessageBox.Information, "Message",
                            "You must input Artist and Album first").exec_()
                return
            self.gimd = google_images_download.googleimagesdownload()
            self._max_count = 20

        query = f"{self.ALBUMARTIST} {self.ALBUM}"
        arguments = {
            "keywords": query,
            "limit": 500,
            "size": "large",
            "no_download": True,
            "no_download_thumbs": True,
            "thumbnail": True
        }

        # init thread that preforms the search
        self.image_search_thread = Thread(target=self.gimd.download,
                                          args=(arguments, ),
                                          name="ImageSearch")
        self.image_search_thread.daemon = True
        self.image_search_thread.start()

        # setup dialog layout and connect to button signals
        self.search_dialog = SearchDialog(query)
        self.search_dialog.pictureClicked.connect(self._select_picture)
        self.search_dialog.load_button.clicked.connect(self._load_more)
        # stop image download on dialog close
        self.search_dialog.finished.connect(self.gimd.close)
        # TODO connect
        self.search_dialog.search_button.clicked.connect(self._do_nothing)
        self.search_dialog.browser_button.clicked.connect(self._do_nothing)

        self.search_dialog.show()

        QTimer.singleShot(0, self._async_loader)

        self.log.info("Searching for Cover Art")

    def _load_more(self):
        """ Method that raises the maximum number of images to download and
        then continues the search.
        """

        # raise the limit of max downloaded images
        self._max_count += 20

        # check if there is enough images to load
        # if not adjust the limit accordingly
        if len(self.images) >= self.gimd.max:
            QMessageBox(QMessageBox.Information, "Message",
                        "No more images to load").exec_()
            self.log.debug("No more images to load")
            return

        if self._max_count > self.gimd.max:
            self._max_count = self.gimd.max

        # continue loading images
        QTimer.singleShot(0, self._async_loader)

    @exception(log)
    def _async_loader(self):
        """ Periodically checks background thread (every 50 ms) for new
        downloaded images. When a new image is found it is loaded and passed to
        GUI dialog to display.

        See also
        --------
        :class:`wiki_music.gui_lib.cover_art.PictureEdit`
            dialog window that displays the downloaded images
        """
        dim: str

        def continue_load(flip: bool = False) -> Tuple[bool, bool]:
            """ Decides if more images should be loaded from background thread.

            Returns
            -------
            Tuple[bool, bool]
                first coresponds to ImageSearch window being visible and the
                other reports if enough images has been loaded
            """

            if flip:
                return [not c for c in continue_load()]
            else:
                return (self.search_dialog.isVisible(),
                        len(self.images) < self._max_count)

        # in one singleShot of the timer we load as mmuch images as available
        while True:
            try:
                image = self.gimd.stack.get(block=False)
            except queue.Empty:
                # if stack is empty break the loop
                break
            else:
                # if item is present try to red it
                try:
                    dim = (f"{image['dim'][1][0]}x{image['dim'][1][1]}\n"
                           f"{(image['dim'][0] / 1024):.2f}Kb")
                except TypeError as e:
                    # if we couldnat load image parameters, don't show it
                    self.log.debug(e)
                    pass
                else:
                    # show and store image
                    self.search_dialog.add_pic(dim, image["thumb"])
                    self.images.append(image)
                    # update progress bar
                    self.search_dialog.progressBar.setValue(
                        int(100 * len(self.images) / self._max_count))

                # indicate that the item is processed
                self.gimd.stack.task_done()

            # don't load more images than allowed
            if any(continue_load(flip=True)):
                break

        # if not enough images were loaded and search window was not destroyed
        # schedule another run
        if all(continue_load()):
            QTimer.singleShot(50, self._async_loader)

    @exception(log)
    def _select_picture(self, index: int):
        """ Method that initializes necessary classes for selected picture
        editing.

        Parameters
        ----------
        index: int
            position of the picture in downloade pictures list
        """

        # 0-th position is size in KB, and 1-st position is tuple of dimensions
        dimensions: Tuple[int, int] = self.images[index]["dim"][1]
        url: str = self.images[index]["url"]

        self.log.info(f"Downloading full size cover art from: {url}")

        # create dialog to handle image editing
        self.picture_editor = PictureEdit(dimensions, get_image(url), self.log,
                                          self.work_dir)
        # if choice was accepted close download, show selected image in gui
        # close search dialog, and save picture
        self.picture_editor.accepted.connect(self.gimd.close)
        self.picture_editor.editingFinished.connect(self._display_image)
        self.picture_editor.accepted.connect(self.search_dialog.close)
        self.picture_editor.accepted.connect(self.picture_editor.picture_save)

        # TODO see discusion in gui utils module
        # clipboard pasting is disabled for now !!!
        self.picture_editor.ca_clipboard.stateChanged.connect(
            self._do_nothing)
        ###########################################
        self.picture_editor.show()



"""Defines aliases for colorama color constrants to be used in whole package.
"""

from colorama import Fore, init

__all__ = ["GREEN", "LGREEN", "RESET", "YELLOW", "LBLUE", "CYAN"]

init(convert=True, autoreset=True)

GREEN: str = Fore.GREEN  #: used to higlight important messages in CLI mode
LGREEN: str = Fore.LIGHTGREEN_EX  #: highlight less important messages in CLI mode
RESET: str = Fore.RESET  #: resets the foreground color to default
YELLOW: str = Fore.YELLOW  #: used to highlight warnings in CLI mode
LBLUE: str = Fore.LIGHTBLUE_EX  #: used to highlight found files in CLI mode
CYAN: str = Fore.CYAN  #: used to designate user input is needed in CLI mode
RED: str = Fore.RED  #: uused to highlight errors in CLI mode



"""Extractor for darklyrics.com."""

import logging
import re
from datetime import datetime

from ..extractor import LyricsExtractor
from ..models.lyrics import Lyrics
from ..models import exceptions

log = logging.getLogger(__name__)


class Darklyrics(LyricsExtractor):
    """Class for extracting lyrics."""

    name = "Darklyrics"
    url = "http://www.darklyrics.com/"
    display_url = "darklyrics.com"

    def extract_lyrics(self, url_data, song, artist):
        """Extract lyrics."""

        bs = url_data.bs

        # get list of album tracks
        track_list = bs.find('div', class_='albumlyrics')

        if not track_list:
            raise exceptions.NoLyrics

        # get table header
        header = track_list.find("h2").text

        # get tracklist
        track_list = track_list.find_all("a", href=re.compile(r"#"))
        track_list = [t.text for t in track_list]

        # convert to dict, delimiter is . or :
        for i, t in enumerate(track_list):
            track_list[i] = re.split(r"\.|:", t)

        # remove whitespaces
        track_list = {t[0]: t[1].strip() for t in track_list}

        # find release date
        for string in re.findall(r'\(.*?\)', header):
            string = re.sub(r"\(|\)", "", string)
            if string.isdigit():
                date_str = string
                break
        else:
            date_str = "9999"

        release_date = datetime.strptime(date_str, "%Y")

        # find the desired song
        for key, value in track_list.items():

            if song.lower() in value.lower():
                song_number = int(key)
                title = track_list[key]

        try:
            lyrics_div = bs.find('div', class_='lyrics')
            # split into separate lyrics
            lyrics = lyrics_div.prettify().split('</h3>')
            lyric = self.process_lyric(lyrics[song_number])
        except Exception as e:
            log.exception(e)

        return Lyrics(title, lyric, artist=artist.title(),
                      release_date=release_date)

    @staticmethod
    def process_lyric(lyric):
        lyric = lyric[:lyric.find('<h3>')]  # remove tail
        # Set linebreaks
        lyric = lyric.replace('<br/>', '')
        # Remove italic
        lyric = lyric.replace('</i>', '')
        lyric = lyric.replace('<i>', '')
        # Remove trailing divs
        lyric = lyric.split('<div')[0]
        result = ''
        split_lyric = lyric.splitlines()
        for line_number in range(len(split_lyric) - 1):
            line = split_lyric[line_number].rstrip().strip()
            next_line = split_lyric[line_number + 1].rstrip()
            last_line = split_lyric[max(line_number - 1, 0)].rstrip()
            # Remove duplicate blank lines
            if line is not '' or (line is '' and next_line is '' and
                                  last_line is not ''):
                result = result + '\n' + line
        return result.strip()

